r"""Guidance and posterior sampling."""

from .cfg import CFGDenoiser  # noqa: F401
from .diffpir import DiffPIRDenoiser  # noqa: F401
from .dps import DPSSampler  # noqa: F401
from .jfps import JFPSDenoiser  # noqa: F401
from .mmps import MMPSDenoiser  # noqa: F401
from .pgdm import PGDMSampler  # noqa: F401
from .repaint import RePaintSampler  # noqa: F401
from .tds import TDSSampler  # noqa: F401
from .tmpd import TMPDenoiser  # noqa: F401



r"""Utilities for debugging."""

__all__ = [
    "RaiseMock",
]

from unittest.mock import Mock


class RaiseMock(Mock):
    r"""Creates an object that raises an error whenever it or its attributes are called.

    Arguments:
        error: The error to be raised.
    """

    def __init__(self, error: Exception, **kwargs) -> None:
        super().__init__(side_effect=error, **kwargs)

    def _get_child_mock(self, **kwargs) -> Mock:
        return super()._get_child_mock(error=self.side_effect, **kwargs)



r"""Tests for the azula.nn.unet module."""

import pytest
import torch

from pathlib import Path

from azula.nn.unet import UNet


@pytest.mark.parametrize("length", [15, 16])
@pytest.mark.parametrize("in_channels, out_channels", [(3, 5)])
@pytest.mark.parametrize("mod_features", [0, 16])
@pytest.mark.parametrize("dropout", [None, 0.1])
@pytest.mark.parametrize("spatial", [1, 2])
@pytest.mark.parametrize("checkpointing", [False, True])
@pytest.mark.parametrize("batch_size", [4])
def test_UNet(
    tmp_path: Path,
    length: int,
    in_channels: int,
    out_channels: int,
    mod_features: int,
    dropout: float,
    spatial: int,
    checkpointing: bool,
    batch_size: int,
) -> None:
    make = lambda: UNet(
        in_channels=in_channels,
        out_channels=out_channels,
        mod_features=mod_features,
        hid_channels=(5, 7, 11),
        hid_blocks=(1, 2, 3),
        dropout=dropout,
        spatial=spatial,
        checkpointing=checkpointing,
    )

    unet = make()
    unet.train()

    # Call
    x = torch.randn((batch_size, in_channels) + (length,) * spatial)
    mod = torch.randn(batch_size, mod_features)
    y = unet(x, mod)

    assert y.ndim == x.ndim
    assert y.shape[0] == batch_size
    assert y.shape[1] == out_channels
    assert y.shape[2:] == x.shape[2:]

    ## Grads
    assert y.requires_grad

    loss = y.square().sum()
    loss.backward()

    for p in unet.parameters():
        assert p.grad is not None
        assert torch.all(torch.isfinite(p.grad))

    # Save
    torch.save(unet.state_dict(), tmp_path / "state.pth")

    # Load
    copy = make()
    copy.load_state_dict(torch.load(tmp_path / "state.pth", weights_only=True))

    unet.eval()
    copy.eval()

    y_unet = unet(x, mod)
    y_copy = copy(x, mod)

    assert torch.allclose(y_unet, y_copy)

    # Float16
    if torch.__version__ < "2.3":
        return

    unet.to(torch.float16)
    y16 = unet(x.to(torch.float16), mod.to(torch.float16))

    unet.to(torch.float32)
    y32 = unet(x.to(torch.float32), mod.to(torch.float32))

    err = (y32 - y16).abs().flatten()

    assert torch.quantile(err, 0.99) < 1e-3
    assert torch.max(err) < 1e-2



# Adapted from
# https://github.com/LTH14/JiT/blob/869190a33f59271724cd1bfddd85777501dadf03/model_jit.py

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from .util import RMSNorm, VisionRotaryEmbeddingFast, get_2d_sincos_pos_embed


def modulate(x, shift, scale):
    return x * (1 + scale.unsqueeze(1)) + shift.unsqueeze(1)


class BottleneckPatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(
        self, img_size=224, patch_size=16, in_chans=3, pca_dim=768, embed_dim=768, bias=True
    ) -> None:
        super().__init__()
        img_size = (img_size, img_size)
        patch_size = (patch_size, patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj1 = nn.Conv2d(
            in_chans, pca_dim, kernel_size=patch_size, stride=patch_size, bias=False
        )
        self.proj2 = nn.Conv2d(pca_dim, embed_dim, kernel_size=1, stride=1, bias=bias)

    def forward(self, x):
        _, _, H, W = x.shape
        assert H == self.img_size[0] and W == self.img_size[1], (
            f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        )
        x = self.proj2(self.proj1(x)).flatten(2).transpose(1, 2)
        return x


class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """

    def __init__(self, hidden_size, frequency_embedding_size=256) -> None:
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """
        # https://github.com/openai/glide-text2im/blob/main/glide_text2im/nn.py
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period)
            * torch.arange(start=0, end=half, dtype=torch.float32, device=t.device)
            / half
        )
        args = t[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1).to(t.dtype)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq)
        return t_emb


class LabelEmbedder(nn.Module):
    """
    Embeds class labels into vector representations. Also handles label dropout for classifier-free guidance.
    """

    def __init__(self, num_classes, hidden_size) -> None:
        super().__init__()
        self.embedding_table = nn.Embedding(num_classes + 1, hidden_size)
        self.num_classes = num_classes

    def forward(self, labels):
        embeddings = self.embedding_table(labels)
        return embeddings


class Attention(nn.Module):
    def __init__(
        self, dim, num_heads=8, qkv_bias=True, qk_norm=True, attn_drop=0.0, proj_drop=0.0
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.q_norm = RMSNorm(head_dim) if qk_norm else nn.Identity()
        self.k_norm = RMSNorm(head_dim) if qk_norm else nn.Identity()

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, rope):
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        q = self.q_norm(q)
        k = self.k_norm(k)

        q = rope(q)
        k = rope(k)

        x = torch.nn.functional.scaled_dot_product_attention(
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0
        )

        x = x.transpose(1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class SwiGLUFFN(nn.Module):
    def __init__(self, dim: int, hidden_dim: int, drop=0.0, bias=True) -> None:
        super().__init__()
        hidden_dim = int(hidden_dim * 2 / 3)
        self.w12 = nn.Linear(dim, 2 * hidden_dim, bias=bias)
        self.w3 = nn.Linear(hidden_dim, dim, bias=bias)
        self.ffn_dropout = nn.Dropout(drop)

    def forward(self, x):
        x12 = self.w12(x)
        x1, x2 = x12.chunk(2, dim=-1)
        hidden = F.silu(x1) * x2
        return self.w3(self.ffn_dropout(hidden))


class FinalLayer(nn.Module):
    """
    The final layer of JiT.
    """

    def __init__(self, hidden_size, patch_size, out_channels) -> None:
        super().__init__()
        self.norm_final = RMSNorm(hidden_size)
        self.linear = nn.Linear(hidden_size, patch_size * patch_size * out_channels, bias=True)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(), nn.Linear(hidden_size, 2 * hidden_size, bias=True)
        )

    @torch.compile
    def forward(self, x, c):
        shift, scale = self.adaLN_modulation(c).chunk(2, dim=1)
        x = modulate(self.norm_final(x), shift, scale)
        x = self.linear(x)
        return x


class JiTBlock(nn.Module):
    def __init__(
        self, hidden_size, num_heads, mlp_ratio=4.0, attn_drop=0.0, proj_drop=0.0
    ) -> None:
        super().__init__()
        self.norm1 = RMSNorm(hidden_size, eps=1e-6)
        self.attn = Attention(
            hidden_size,
            num_heads=num_heads,
            qkv_bias=True,
            qk_norm=True,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
        )
        self.norm2 = RMSNorm(hidden_size, eps=1e-6)
        mlp_hidden_dim = int(hidden_size * mlp_ratio)
        self.mlp = SwiGLUFFN(hidden_size, mlp_hidden_dim, drop=proj_drop)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(), nn.Linear(hidden_size, 6 * hidden_size, bias=True)
        )

    @torch.compile
    def forward(self, x, c, feat_rope=None):
        shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp = self.adaLN_modulation(
            c
        ).chunk(6, dim=-1)
        x = x + gate_msa.unsqueeze(1) * self.attn(
            modulate(self.norm1(x), shift_msa, scale_msa), rope=feat_rope
        )
        x = x + gate_mlp.unsqueeze(1) * self.mlp(modulate(self.norm2(x), shift_mlp, scale_mlp))
        return x


class JiT(nn.Module):
    """
    Just image Transformer.
    """

    def __init__(
        self,
        input_size=256,
        patch_size=16,
        in_channels=3,
        hidden_size=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4.0,
        attn_drop=0.0,
        proj_drop=0.0,
        num_classes=1000,
        bottleneck_dim=128,
        in_context_len=32,
        in_context_start=8,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.in_context_len = in_context_len
        self.in_context_start = in_context_start
        self.num_classes = num_classes

        # time and class embed
        self.t_embedder = TimestepEmbedder(hidden_size)
        self.y_embedder = LabelEmbedder(num_classes, hidden_size)

        # linear embed
        self.x_embedder = BottleneckPatchEmbed(
            input_size, patch_size, in_channels, bottleneck_dim, hidden_size, bias=True
        )

        # use fixed sin-cos embedding
        num_patches = self.x_embedder.num_patches
        self.pos_embed = nn.Parameter(
            torch.zeros(1, num_patches, hidden_size), requires_grad=False
        )

        # in-context cls token
        if self.in_context_len > 0:
            self.in_context_posemb = nn.Parameter(
                torch.zeros(1, self.in_context_len, hidden_size), requires_grad=True
            )
            torch.nn.init.normal_(self.in_context_posemb, std=0.02)

        # rope
        half_head_dim = hidden_size // num_heads // 2
        hw_seq_len = input_size // patch_size
        self.feat_rope = VisionRotaryEmbeddingFast(
            dim=half_head_dim, pt_seq_len=hw_seq_len, num_cls_token=0
        )
        self.feat_rope_incontext = VisionRotaryEmbeddingFast(
            dim=half_head_dim, pt_seq_len=hw_seq_len, num_cls_token=self.in_context_len
        )

        # transformer
        self.blocks = nn.ModuleList([
            JiTBlock(
                hidden_size,
                num_heads,
                mlp_ratio=mlp_ratio,
                attn_drop=attn_drop if (depth // 4 * 3 > i >= depth // 4) else 0.0,
                proj_drop=proj_drop if (depth // 4 * 3 > i >= depth // 4) else 0.0,
            )
            for i in range(depth)
        ])

        # linear predict
        self.final_layer = FinalLayer(hidden_size, patch_size, self.out_channels)

        self.initialize_weights()

    def initialize_weights(self) -> None:
        # Initialize transformer layers:
        def _basic_init(module) -> None:
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)

        self.apply(_basic_init)

        # Initialize (and freeze) pos_embed by sin-cos embedding:
        pos_embed = get_2d_sincos_pos_embed(
            self.pos_embed.shape[-1], int(self.x_embedder.num_patches**0.5)
        )
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

        # Initialize patch_embed like nn.Linear (instead of nn.Conv2d):
        w1 = self.x_embedder.proj1.weight.data
        nn.init.xavier_uniform_(w1.view([w1.shape[0], -1]))
        w2 = self.x_embedder.proj2.weight.data
        nn.init.xavier_uniform_(w2.view([w2.shape[0], -1]))
        nn.init.constant_(self.x_embedder.proj2.bias, 0)

        # Initialize label embedding table:
        nn.init.normal_(self.y_embedder.embedding_table.weight, std=0.02)

        nn.init.normal_(self.t_embedder.mlp[0].weight, std=0.02)
        nn.init.normal_(self.t_embedder.mlp[2].weight, std=0.02)

        # Zero-out adaLN modulation layers:
        for block in self.blocks:
            nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(block.adaLN_modulation[-1].bias, 0)

        # Zero-out output layers:
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].bias, 0)

        nn.init.constant_(self.final_layer.linear.weight, 0)
        nn.init.constant_(self.final_layer.linear.bias, 0)

    def unpatchify(self, x, p):
        """
        x: (N, T, patch_size**2 * C)
        imgs: (N, H, W, C)
        """
        c = self.out_channels
        h = w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, c))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], c, h * p, h * p))
        return imgs

    def forward(self, x, t, y):
        """
        x: (N, C, H, W)
        t: (N,)
        y: (N,)
        """
        # class and time embeddings
        t_emb = self.t_embedder(t)
        y_emb = self.y_embedder(y)
        c = t_emb + y_emb

        # forward JiT
        x = self.x_embedder(x)
        x += self.pos_embed

        for i, block in enumerate(self.blocks):
            # in-context
            if self.in_context_len > 0 and i == self.in_context_start:
                in_context_tokens = y_emb.unsqueeze(1).repeat(1, self.in_context_len, 1)
                in_context_tokens += self.in_context_posemb
                x = torch.cat([in_context_tokens, x], dim=1)
            x = block(
                x, c, self.feat_rope if i < self.in_context_start else self.feat_rope_incontext
            )

        x = x[:, self.in_context_len :]

        x = self.final_layer(x, c)
        output = self.unpatchify(x, self.patch_size)

        return output


def JiT_B_16(**kwargs):
    return JiT(
        depth=12,
        hidden_size=768,
        num_heads=12,
        bottleneck_dim=128,
        in_context_len=32,
        in_context_start=4,
        patch_size=16,
        **kwargs,
    )


def JiT_B_32(**kwargs):
    return JiT(
        depth=12,
        hidden_size=768,
        num_heads=12,
        bottleneck_dim=128,
        in_context_len=32,
        in_context_start=4,
        patch_size=32,
        **kwargs,
    )


def JiT_L_16(**kwargs):
    return JiT(
        depth=24,
        hidden_size=1024,
        num_heads=16,
        bottleneck_dim=128,
        in_context_len=32,
        in_context_start=8,
        patch_size=16,
        **kwargs,
    )


def JiT_L_32(**kwargs):
    return JiT(
        depth=24,
        hidden_size=1024,
        num_heads=16,
        bottleneck_dim=128,
        in_context_len=32,
        in_context_start=8,
        patch_size=32,
        **kwargs,
    )


def JiT_H_16(**kwargs):
    return JiT(
        depth=32,
        hidden_size=1280,
        num_heads=16,
        bottleneck_dim=256,
        in_context_len=32,
        in_context_start=10,
        patch_size=16,
        **kwargs,
    )


def JiT_H_32(**kwargs):
    return JiT(
        depth=32,
        hidden_size=1280,
        num_heads=16,
        bottleneck_dim=256,
        in_context_len=32,
        in_context_start=10,
        patch_size=32,
        **kwargs,
    )


JiT_models = {
    "JiT-B/16": JiT_B_16,
    "JiT-B/32": JiT_B_32,
    "JiT-L/16": JiT_L_16,
    "JiT-L/32": JiT_L_32,
    "JiT-H/16": JiT_H_16,
    "JiT-H/32": JiT_H_32,
}



r"""Twisted Posterior Sampling (TDS) internals.

References:
    | Practical and Asymptotically Exact Conditional Sampling in Diffusion Models (Wu et al., 2023)
    | https://arxiv.org/abs/2306.17775
"""

__all__ = [
    "TDSSampler",
]

import torch

from collections.abc import Callable
from einops import reduce
from torch import Tensor
from torch.distributions import Normal

from ..denoise import Denoiser
from ..sample import Sampler


class TDSSampler(Sampler):
    r"""Creates a TDS sampler.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        twist: A twisting function $\log p(y | \hat{x}, t)$.
        kwargs: Keyword arguments passed to :class:`azula.sample.Sampler`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        twist: Callable[[Tensor], Tensor],
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.twist = twist

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        r"""Simulates the reverse process from :math:`t_T` to :math:`t_0`.

        Arguments:
            x: A set of :math:`K` noisy tensors :math:`x_{t_T}`, with shape :math:`(K, *)`.
            kwargs: Optional keyword arguments.

        Returns:
            The clean(er) tensors :math:`x_{t_0}`, with shape :math:`(K, *)`.
        """

        return super().__call__(x, carry={}, **kwargs)

    @torch.no_grad()
    def step(self, x_t: Tensor, t: Tensor, s: Tensor, carry: dict, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        with torch.enable_grad():
            x_t = x_t.detach().requires_grad_()
            x_hat = self.denoiser(x_t, t, **kwargs).mean

            log_p_y = self.twist(x_hat, sigma_t / alpha_t)
            score_y = torch.autograd.grad(log_p_y.sum(), x_t)[0]

        # Resample
        log_p_y = reduce(log_p_y, "K ... -> K", "sum")

        if "log_w" in carry:
            log_w = log_p_y + carry["log_w"]
        else:
            log_w = log_p_y

        w = torch.softmax(log_w, dim=0)
        k = torch.multinomial(w, len(w), replacement=True)

        x_t, x_hat, log_p_y, score_y = x_t[k], x_hat[k], log_p_y[k], score_y[k]

        # Proposal
        def ddpm(x_t: Tensor, x: Tensor) -> Normal:
            eps = (x_t - alpha_t * x) / sigma_t
            tau = (alpha_t / alpha_s * sigma_s / sigma_t) ** 2

            return Normal(
                loc=alpha_s * x + sigma_s * torch.sqrt(tau) * eps,
                scale=sigma_s * torch.sqrt(1 - tau),
                validate_args=False,
            )

        q_s = ddpm(x_t, x_hat)
        q_s_y = ddpm(x_t, x_hat + sigma_t**2 / alpha_t * score_y)

        x_s = q_s_y.sample()

        # Reweight
        log_q_xs = reduce(q_s.log_prob(x_s), "K ... -> K", "sum")
        log_q_xs_y = reduce(q_s_y.log_prob(x_s), "K ... -> K", "sum")

        carry["log_w"] = log_q_xs - log_q_xs_y - log_p_y

        return x_s



r"""Moment Matching Posterior Sampling (MMPS) internals.

References:
    | Learning Diffusion Priors from Observations by Expectation Maximization (Rozet et al., 2024)
    | https://arxiv.org/abs/2405.13712
"""

__all__ = [
    "MMPSDenoiser",
]

import torch

from collections.abc import Callable
from functools import partial
from torch import Tensor

from ..denoise import Denoiser, DiracPosterior
from ..linalg.covariance import Covariance, DiagonalCovariance
from ..linalg.solve import cg, gmres
from ..noise import Schedule


class MMPSDenoiser(Denoiser):
    r"""Creates a MMPS denoiser module.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y \sim \mathcal{N}(A(x), \Sigma_y)`, with shape :math:`(*, D)`.
        A: The forward operator :math:`x \mapsto A(x)`.
        cov_y: The noise covariance :math:`\Sigma_y`. If `cov_y` is a tensor, it is
            assumed to be the variance :math:`\sigma_y^2` and :math:`\Sigma_y =
            \mathrm{diag}(\sigma_y^2)`.
        solver: The linear solver name (:py:`"cg"` or :py:`"gmres"`).
        iterations: The number of solver iterations.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        A: Callable[[Tensor], Tensor],
        cov_y: Tensor | Covariance,
        solver: str = "gmres",
        iterations: int = 1,
    ) -> None:
        super().__init__()

        self.denoiser = denoiser

        self.y = y
        self.A = A

        if isinstance(cov_y, Covariance):
            self.cov_y = cov_y
        else:
            self.cov_y = DiagonalCovariance(cov_y)

        if solver == "cg":
            self.solve = partial(cg, iterations=iterations)
        elif solver == "gmres":
            self.solve = partial(gmres, iterations=iterations)
        else:
            raise ValueError(f"Unknown solver '{solver}'.")

    @property
    def schedule(self) -> Schedule:
        return self.denoiser.schedule

    @torch.no_grad()
    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        alpha_t, sigma_t = self.schedule(t)
        gamma_t = sigma_t**2 / alpha_t

        with torch.enable_grad():
            x_t = x_t.detach().requires_grad_()
            q = self.denoiser(x_t, t, **kwargs)

            x_hat = q.mean
            y_hat = self.A(x_hat)

        def A(v: Tensor) -> Tensor:
            return torch.func.jvp(self.A, (x_hat.detach(),), (v,))[1]

        def At(v: Tensor) -> Tensor:
            return torch.autograd.grad(y_hat, x_hat, v, retain_graph=True)[0]

        def cov_x(v: Tensor) -> Tensor:
            return gamma_t * torch.autograd.grad(x_hat, x_t, v, retain_graph=True)[0]

        def cov_y(v: Tensor) -> Tensor:
            return self.cov_y(v) + A(cov_x(At(v)))

        grad = self.y - y_hat
        grad = self.solve(A=cov_y, b=grad)
        grad = gamma_t * torch.autograd.grad(y_hat, x_t, grad)[0]

        return DiracPosterior(mean=x_hat + grad)



r"""Just Image Transformer (JIT) plugin.

.. code-block:: python

    from azula.plugins import jit

References:
    | Back to Basics: Let Denoising Generative Models Denoise (Li et al., 2025)
    | https://arxiv.org/abs/2511.13720
"""

__all__ = [
    "JITDenoiser",
    "load_model",
]

import os
import torch
import torch.nn as nn

from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.hub import download
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import RectifiedSchedule, Schedule

from ._src.model import JiT_models
from ..utils import load_cards


class JITDenoiser(Denoiser):
    r"""Creates a JIT denoiser.

    Arguments:
        backbone: A time conditional network.
        schedule: A noise schedule. If :py:`None`, use
            :class:`azula.noise.RectifiedSchedule` instead.
        num_classes: The number of classes.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
        num_classes: int = 1000,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = RectifiedSchedule()
        else:
            self.schedule = schedule

        self.num_classes = num_classes

    def forward(
        self,
        x_t: Tensor,
        t: Tensor,
        label: Tensor | None = None,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, 3, H, W)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            label: The class label :math:`c` as an integer, with shape :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \mu_\phi(x_t \mid c))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = 1 / (alpha_t + sigma_t)
        c_time = (alpha_t / (alpha_t + sigma_t)).flatten()

        B, _, _, _ = x_t.shape

        dtype = get_module_dtype(self.backbone)

        if label is None:
            label = torch.as_tensor(self.num_classes, device=x_t.device)

        output = self.backbone(
            (c_in * x_t).to(dtype),
            c_time.to(dtype),
            y=label.expand(B),
            **kwargs,
        ).to(x_t)

        mean = output

        return DiracPosterior(mean=mean)


def load_model(
    name: str,
    ema: bool = True,
    **kwargs,
) -> Denoiser:
    r"""Loads a pre-trained JIT denoiser.

    Arguments:
        name: The pre-trained model name.
        ema: Whether to load EMA weights or not.
        kwargs: Keyword arguments passed to :func:`torch.load`.

    Returns:
        A pre-trained denoiser.
    """

    kwargs.setdefault("map_location", "cpu")
    kwargs.setdefault("weights_only", True)

    card = load_cards(__name__)[name]

    state = torch.load(
        os.path.join(
            download(card.url, hash_prefix=card.hash, extract=True),
            "checkpoint-last.pth",
        ),
        **kwargs,
    )

    if ema:
        state = state["model_ema1"]
    else:
        state = state["model"]

    state = {k.removeprefix("net."): v for k, v in state.items()}

    with skip_init():
        denoiser = make_model(**card.config)

    denoiser.backbone.load_state_dict(state)

    return denoiser.eval()


def make_model(model: str = "JiT-B/16", **kwargs) -> Denoiser:
    r"""Initializes a JIT denoiser."""

    backbone = JiT_models[model](**kwargs)

    return JITDenoiser(
        backbone,
        num_classes=backbone.num_classes,
    )



r"""RePaint internals.

References:
    | RePaint: Inpainting using Denoising Diffusion Probabilistic Models (Lugmayr et al., 2022)
    | https://arxiv.org/abs/2201.09865
"""

__all__ = [
    "RePaintSampler",
]

import torch

from torch import BoolTensor, Tensor

from ..denoise import Denoiser
from ..sample import DDIMSampler


class RePaintSampler(DDIMSampler):
    r"""Creates a RePaint sampler.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y = m \odot x`.
        mask: The observation mask :math:`m`.
        iterations: The number of RePaint iterations per step.
        kwargs: Keyword arguments passed to :class:`azula.sample.DDIMSampler`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        mask: BoolTensor,
        iterations: int = 3,
        **kwargs,
    ) -> None:
        super().__init__(denoiser, **kwargs)

        self.y = y
        self.mask = mask

        self.iterations = iterations

    @torch.no_grad()
    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        for _ in range(self.iterations):
            x_s = super().step(x_t, t, s, **kwargs)
            x_s = torch.where(
                self.mask,
                alpha_s * self.y + sigma_s * torch.randn_like(self.y),
                x_s,
            )

            x_t = alpha_t / alpha_s * x_s + alpha_t * torch.sqrt(
                (sigma_t / alpha_t) ** 2 - (sigma_s / alpha_s) ** 2
            ) * torch.randn_like(x_s)

        return x_s



r"""Tests for the azula.linalg.solve module."""

import pytest
import torch

from collections.abc import Iterator, Sequence
from functools import partial

from azula.linalg.solve import cg, gmres


@pytest.fixture(autouse=True, scope="module")
def torch_float64() -> Iterator[None]:
    default = torch.get_default_dtype()

    try:
        yield torch.set_default_dtype(torch.float64)
    finally:
        torch.set_default_dtype(default)


@pytest.mark.parametrize("rank", [3, 5])
@pytest.mark.parametrize("batch", [(), (64,)])
def test_cg(rank: int, batch: Sequence[int]) -> None:
    U = torch.randn(*batch, 5, rank)
    A = partial(torch.einsum, "...ij,...j", U @ U.mT)

    x = torch.randn(*batch, 5)
    Ax = A(x)

    # x_0 = 0
    y = cg(A=A, b=Ax, iterations=rank)
    Ay = A(y)

    assert y.shape == x.shape
    assert torch.allclose(Ay, Ax, atol=1e-6)

    # x_0 = x
    y = cg(A=A, b=Ax, x0=x, iterations=1)
    Ay = A(y)

    assert y.shape == x.shape
    assert torch.allclose(Ay, Ax)


@pytest.mark.parametrize("rank", [3, 5])
@pytest.mark.parametrize("batch", [(), (64,)])
def test_gmres(rank: int, batch: Sequence[int]) -> None:
    U = torch.randn(*batch, 5, rank)
    V = torch.randn(*batch, rank, 5)
    A = partial(torch.einsum, "...ij,...j", U @ V)

    x = torch.randn(*batch, 5)
    Ax = A(x)

    # x_0 = 0
    y = gmres(A=A, b=Ax, iterations=rank)
    Ay = A(y)

    assert y.shape == x.shape
    assert torch.allclose(Ay, Ax, atol=1e-6)

    # x_0 = x
    y = gmres(A=A, b=Ax, x0=x, iterations=1)
    Ay = A(y)

    assert y.shape == x.shape
    assert torch.allclose(Ay, Ax)



r"""Jacobian-Free Posterior Sampling (JFPS) internals."""

from __future__ import annotations

__all__ = [
    "JFPSDenoiser",
]

import torch

from collections.abc import Callable
from functools import partial
from torch import Tensor
from typing import Literal

from ..denoise import Denoiser, DiracPosterior
from ..linalg.covariance import Covariance, IsotropicCovariance
from ..linalg.solve import cg, gmres
from ..noise import Schedule


class JFPSDenoiser(Denoiser):
    r"""Creates a JFPS denoiser module.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y \sim \mathcal{N}(A(x), \Sigma_y)`, with shape :math:`(*, D)`.
        A: The forward operator :math:`x \mapsto A(x)`.
        cov_y: The noise covariance :math:`\Sigma_y`.
        cov_x: The signal covariance :math:`\Sigma_x`.
        solver: The linear solver name.
        iterations: The number of solver iterations.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        A: Callable[[Tensor], Tensor],
        cov_y: Covariance,
        cov_x: Covariance,
        solver: Literal["cg", "gmres"] = "cg",
        iterations: int = 1,
    ) -> None:
        super().__init__()

        self.denoiser = denoiser

        self.y = y
        self.A = A
        self.cov_y = cov_y
        self.cov_x = cov_x

        if solver == "cg":
            self.solve = partial(cg, iterations=iterations)
        elif solver == "gmres":
            self.solve = partial(gmres, iterations=iterations)
        else:
            raise ValueError(f"Unknown solver '{solver}'.")

    @property
    def schedule(self) -> Schedule:
        return self.denoiser.schedule

    @torch.no_grad()
    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        alpha_t, sigma_t = self.schedule(t)

        q = self.denoiser(x_t, t, **kwargs)

        with torch.enable_grad():
            x_hat = q.mean.detach().requires_grad_()
            y_hat = self.A(x_hat)

        def A(v: Tensor) -> Tensor:
            return torch.func.jvp(self.A, (x_hat.detach(),), (v,))[1]

        def At(v: Tensor) -> Tensor:
            return torch.autograd.grad(y_hat, x_hat, v, retain_graph=True)[0]

        cov_t = IsotropicCovariance(sigma_t**2 / alpha_t**2)
        cov_x = (self.cov_x.inv + cov_t.inv).inv

        def cov_y(v: Tensor) -> Tensor:
            return self.cov_y(v) + A(cov_x(At(v)))

        grad = self.y - y_hat
        grad = self.solve(A=cov_y, b=grad)
        grad = torch.autograd.grad(y_hat, x_hat, grad)[0]
        grad = cov_x(grad)

        return DiracPosterior(mean=x_hat + grad)



r"""Tests for the azula.linalg.solve module."""

import pytest
import torch

from collections.abc import Iterator, Sequence

from azula.linalg.covariance import (
    DiagonalCovariance,
    DPLRCovariance,
    FullCovariance,
    IsotropicCovariance,
    KroneckerCovariance,
)


@pytest.fixture(autouse=True, scope="module")
def torch_float64() -> Iterator[None]:
    default = torch.get_default_dtype()

    try:
        yield torch.set_default_dtype(torch.float64)
    finally:
        torch.set_default_dtype(default)


@pytest.mark.parametrize(
    "covariance_cls",
    [
        IsotropicCovariance,
        DiagonalCovariance,
        FullCovariance,
        DPLRCovariance,
        KroneckerCovariance,
    ],
)
@pytest.mark.parametrize("shape", [(5,), (3, 5)])
@pytest.mark.parametrize("batch", [(), (256,), (16, 16)])
def test_covariances(covariance_cls: type, shape: Sequence[int], batch: Sequence[int]) -> None:
    X = torch.randn(1024, *shape)

    cov = covariance_cls.from_data(X)
    cov_inv = cov.inv

    x = torch.randn(*batch, *shape)

    assert x.shape == cov(x).shape
    assert x.shape == cov_inv(x).shape

    assert torch.allclose(x, cov_inv(cov(x)))
    assert torch.allclose(x, cov(cov_inv(x)))



r"""Common layers."""

__all__ = [
    "ConvNd",
    "LayerNorm",
    "Patchify",
    "RMSNorm",
    "ReLU2",
    "SineEncoding",
    "SwiGLU",
    "Unpatchify",
]

import math
import string
import torch

from collections.abc import Sequence
from einops.layers.torch import Rearrange
from torch import Tensor

from .utils import promote_dtype


def ConvNd(
    in_channels: int,
    out_channels: int,
    spatial: int = 2,
    identity_init: bool = False,
    **kwargs,
) -> torch.nn.Module:
    r"""Returns an N-dimensional convolutional layer.

    Arguments:
        in_channels: The number of input channels :math:`C_i`.
        out_channels: The number of output channels :math:`C_o`.
        spatial: The number of spatial dimensions :math:`N`.
        identity_init: Initialize the convolution as a (pseudo-)identity.
        kwargs: Keyword arguments passed to :class:`torch.nn.Conv2d`.
    """

    CONVS = {
        0: torch.nn.Linear,
        1: torch.nn.Conv1d,
        2: torch.nn.Conv2d,
        3: torch.nn.Conv3d,
    }

    if spatial in CONVS:
        Conv = CONVS[spatial]
    else:
        raise NotImplementedError()

    conv = Conv(in_channels, out_channels, **kwargs)

    if identity_init:
        kernel_size = conv.weight.shape[2:]
        kernel_center = [k // 2 for k in kernel_size]

        eye = torch.zeros_like(conv.weight.data[:in_channels])

        for i in range(min(in_channels, out_channels)):
            eye[(i, i, *kernel_center)] = 1.0

        conv.weight.data[:in_channels].mul_(1e-2)
        conv.weight.data[:in_channels].add_(eye)

    return conv


class ReLU2(torch.nn.Module):
    r"""Creates a ReLU² activation layer.

    .. math:: y = \max(x, 0)^2

    References:
        | Primer: Searching for Efficient Transformers for Language Modeling (So et al., 2021)
        | https://arxiv.org/abs/2109.08668
    """

    def forward(self, x: Tensor) -> Tensor:
        return relu2(x)


def relu2(x: Tensor, /) -> Tensor:
    return torch.nn.functional.relu(x).square()


class SwiGLU(torch.nn.Module):
    r"""Creates a SwiGLU activation layer.

    .. math:: y = x_1 \times x_2 \times \sigma(x_2)

    References:
        | GLU Variants Improve Transformer (Shazeer, 2020)
        | https://arxiv.org/abs/2002.05202
    """

    def forward(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, 2C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, C)`.
        """

        return swiglu(x)


def swiglu(x: Tensor, /) -> Tensor:
    x = x.unflatten(-1, (-1, 2))
    x1, x2 = x[..., 0], x[..., 1]
    return x1 * torch.nn.functional.silu(x2)


class LayerNorm(torch.nn.Module):
    r"""Creates a layer that standardizes features along a dimension.

    .. math:: y = \frac{x - \mathbb{E}[x]}{\sqrt{\mathbb{V}[x] + \epsilon}}

    References:
       | Layer Normalization (Lei Ba et al., 2016)
       | https://arxiv.org/abs/1607.06450

    Arguments:
        dim: The dimension(s) to standardize.
        eps: A numerical stability term.
    """

    def __init__(self, dim: int | Sequence[int], eps: float = 1e-5) -> None:
        super().__init__()

        self.dim = dim
        self.eps = eps

    def extra_repr(self) -> str:
        return f"dim={self.dim}"

    def forward(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:(*).

        Returns:
            The standardized tensor :math:`y`, with shape :math:`(*)`.
        """

        return layer_norm(x, dim=self.dim, eps=self.eps)


@promote_dtype
def layer_norm(x: Tensor, /, dim: int = -1, eps: float = 1e-5) -> Tensor:
    v, m = torch.var_mean(x, dim=dim, keepdim=True)
    return (x - m) * torch.rsqrt(v + eps)


class RMSNorm(torch.nn.Module):
    r"""Creates a layer that normalizes features along a dimension.

    .. math:: y = \frac{x}{\sqrt{\mathbb{E}[x^2] + \epsilon}}

    References:
       | Root Mean Square Layer Normalization (Zhang et al., 2019)
       | https://arxiv.org/abs/1910.07467

    Arguments:
        dim: The dimension(s) to normalize.
        eps: A numerical stability term.
    """

    def __init__(self, dim: int | Sequence[int], eps: float = 1e-5) -> None:
        super().__init__()

        self.dim = dim
        self.eps = eps

    def extra_repr(self) -> str:
        return f"dim={self.dim}"

    def forward(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:(*).

        Returns:
            The normalized tensor :math:`y`, with shape :math:`(*)`.
        """

        return rms_norm(x, dim=self.dim, eps=self.eps)


@promote_dtype
def rms_norm(x: Tensor, /, dim: int = -1, eps: float = 1e-5) -> Tensor:
    return x * torch.rsqrt(torch.mean(torch.square(x), dim=dim, keepdim=True) + eps)


def Patchify(patch_shape: Sequence[int], channel_last: bool = False) -> Rearrange:
    r"""Returns a patch-to-channel layer.

    Arguments:
        patch_shape: The patch shape.
        channel_last: Whether the output channel dimension is first or last.
    """

    ndim = len(patch_shape)

    ABC = string.ascii_uppercase[:ndim]
    abc = string.ascii_lowercase[:ndim]

    in_shape = (f"({A} {a})" for A, a in zip(ABC, abc, strict=True))
    in_shape = "... Z " + " ".join(in_shape)

    if channel_last:
        out_shape = "... " + " ".join(ABC) + " (Z " + " ".join(abc) + ")"
    else:
        out_shape = "... (Z " + " ".join(abc) + ") " + " ".join(ABC)

    lengths = {a: size for a, size in zip(abc, patch_shape, strict=True)}

    return Rearrange(f"{in_shape} -> {out_shape}", **lengths)


def Unpatchify(patch_shape: Sequence[int], channel_last: bool = False) -> Rearrange:
    r"""Returns a channel-to-patch layer.

    Arguments:
        patch_shape: The patch shape.
        channel_last: Whether the input channel dimension is first or last.
    """

    ndim = len(patch_shape)

    ABC = string.ascii_uppercase[:ndim]
    abc = string.ascii_lowercase[:ndim]

    in_shape = (f"({A} {a})" for A, a in zip(ABC, abc, strict=True))
    in_shape = "... Z " + " ".join(in_shape)

    if channel_last:
        out_shape = "... " + " ".join(ABC) + " (Z " + " ".join(abc) + ")"
    else:
        out_shape = "... (Z " + " ".join(abc) + ") " + " ".join(ABC)

    lengths = {a: size for a, size in zip(abc, patch_shape, strict=True)}

    return Rearrange(f"{out_shape} -> {in_shape}", **lengths)


class SineEncoding(torch.nn.Module):
    r"""Creates a sinusoidal positional encoding.

    .. math::
        e_{2i} & = \sin \left( x \times \omega^\frac{-2i}{D} \right) \\
        e_{2i+1} & = \cos \left( x \times \omega^\frac{-2i}{D} \right)

    References:
        | Attention Is All You Need (Vaswani et al., 2017)
        | https://arxiv.org/abs/1706.03762

    Arguments:
        features: The number of embedding features :math:`D`. Must be even.
        omega: The maximum frequency :math:`\omega`.
    """

    def __init__(self, features: int, omega: float = 1e4) -> None:
        super().__init__()

        assert features % 2 == 0

        self.features = features
        self.omega = omega

    def forward(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: The position :math:`x`, with shape :math:`(*)`.

        Returns:
            The embedding vector :math:`e`, with shape :math:`(*, D)`.
        """

        return sine_encoding(x, features=self.features, omega=self.omega)


@promote_dtype
def sine_encoding(x: Tensor, /, features: int, omega: float = 1e4) -> Tensor:
    x = x.unsqueeze(dim=-1)

    freqs = torch.linspace(0, 1, features // 2, dtype=x.dtype, device=x.device)
    freqs = torch.exp(math.log(1 / omega) * freqs)

    return torch.cat(
        (
            torch.sin(x * freqs),
            torch.cos(x * freqs),
        ),
        dim=-1,
    )



r"""Tests for the azula.sample module."""

import pytest
import torch

from collections.abc import Sequence
from functools import partial
from torch import Tensor

from azula.denoise import KarrasDenoiser
from azula.nn.layers import SineEncoding
from azula.noise import VPSchedule
from azula.sample import (
    DDIMSampler,
    DDPMSampler,
    EulerSampler,
    HeunSampler,
    ItoSampler,
    PCSampler,
    REABSampler,
    vABSampler,
    xEABSampler,
    zABSampler,
    zEABSampler,
)


class Dummy(torch.nn.Module):
    def __init__(self, features: int = 5, with_label: bool = False) -> None:
        super().__init__()

        self.with_label = with_label

        self.l1 = torch.nn.Linear(features, 64)
        self.l2 = torch.nn.Linear(64, features)
        self.relu = torch.nn.ReLU()

        self.time_encoding = SineEncoding(64)

    def forward(self, x_t: Tensor, t: Tensor, label: str | None = None) -> Tensor:
        y = self.l1(x_t)
        y = y + self.time_encoding(t)
        y = self.relu(y)
        y = self.l2(y)

        if self.with_label:
            assert isinstance(label, str)
        else:
            assert label is None

        return y


@pytest.mark.parametrize("with_label", [False, True])
@pytest.mark.parametrize("batch", [(), (64,)])
@pytest.mark.parametrize("channels", [5])
def test_samplers(with_label: bool, batch: Sequence[int], channels: int) -> None:
    denoiser = KarrasDenoiser(
        backbone=Dummy(channels, with_label),
        schedule=VPSchedule(),
    )

    Ss = [
        partial(DDPMSampler),
        partial(DDIMSampler, eta=0.0),
        partial(DDIMSampler, eta=1.0),
        partial(EulerSampler),
        partial(HeunSampler),
        partial(ItoSampler, eta=1.0),
        partial(zABSampler),
        partial(vABSampler),
        partial(zEABSampler),
        partial(xEABSampler),
        partial(REABSampler),
        partial(PCSampler, corrections=1),
    ]

    for S in Ss:
        sampler = S(denoiser, steps=64)

        x1 = sampler.init((*batch, channels))

        assert x1.shape == (*batch, channels), S
        assert torch.all(torch.isfinite(x1)), S

        if with_label:
            x0 = sampler(x1, label="cat")
        else:
            x0 = sampler(x1)

        assert x0.shape == (*batch, channels), S
        assert torch.all(torch.isfinite(x0)), S



r"""Tests for the azula.nn.vit module."""

import pytest
import torch

from pathlib import Path

from azula.nn.vit import ViT


@pytest.mark.parametrize("length", [16])
@pytest.mark.parametrize("in_channels, out_channels", [(3, 5)])
@pytest.mark.parametrize("mod_features", [0, 16])
@pytest.mark.parametrize("attention_heads", [4])
@pytest.mark.parametrize("dropout", [None, 0.1])
@pytest.mark.parametrize("spatial", [1, 2])
@pytest.mark.parametrize("rope", [False, True])
@pytest.mark.parametrize("checkpointing", [False, True])
@pytest.mark.parametrize("batch_size", [4])
def test_ViT(
    tmp_path: Path,
    length: int,
    in_channels: int,
    out_channels: int,
    mod_features: int,
    attention_heads: int,
    dropout: float,
    spatial: int,
    rope: bool,
    checkpointing: bool,
    batch_size: int,
) -> None:
    make = lambda: ViT(
        in_channels=in_channels,
        out_channels=out_channels,
        mod_features=mod_features,
        hid_channels=16,
        hid_blocks=3,
        attention_heads=attention_heads,
        dropout=dropout,
        spatial=spatial,
        patch_size=[4] * spatial,
        rope=rope,
        checkpointing=checkpointing,
    )

    vit = make()
    vit.train()

    # Call
    x = torch.randn((batch_size, in_channels) + (length,) * spatial)
    mod = torch.randn(batch_size, mod_features)
    y = vit(x, mod)

    assert y.ndim == x.ndim
    assert y.shape[0] == batch_size
    assert y.shape[1] == out_channels
    assert y.shape[2:] == x.shape[2:]

    ## Grads
    assert y.requires_grad

    loss = y.square().sum()
    loss.backward()

    for p in vit.parameters():
        assert p.grad is not None
        assert torch.all(torch.isfinite(p.grad))

    # Save
    torch.save(vit.state_dict(), tmp_path / "state.pth")

    # Load
    copy = make()
    copy.load_state_dict(torch.load(tmp_path / "state.pth", weights_only=True))

    vit.eval()
    copy.eval()

    y = vit(x, mod)
    y_copy = copy(x, mod)

    assert torch.allclose(y, y_copy)

    # Float16
    if torch.__version__ < "2.3":
        return

    vit.to(torch.float16)
    y16 = vit(x.to(torch.float16), mod.to(torch.float16))

    vit.to(torch.float32)
    y32 = vit(x.to(torch.float32), mod.to(torch.float32))

    err = (y32 - y16).abs().flatten()

    assert torch.quantile(err, 0.99) < 1e-3
    assert torch.max(err) < 1e-2



r"""Sana plugin.

This plugin depends on :mod:`diffusers` and :mod:`transformers`. To use it,
install the dependencies in your environment

.. code-block:: console

    pip install diffusers transformers accelerate

before importing the plugin.

.. code-block:: python

    from azula.plugins import sana

References:
    | SANA: Efficient High-Resolution Image Synthesis with Linear Diffusion Transformers (Xie et al., 2024)
    | https://arxiv.org/abs/2410.10629
"""

__all__ = [
    "AutoEncoder",
    "TextEncoder",
    "SanaDenoiser",
    "load_model",
]

import torch.nn as nn

from collections.abc import Sequence
from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import DecaySchedule, Schedule

from ..utils import load_cards, patch_diffusers, patch_mmap_safetensors


class AutoEncoder(nn.Module):
    r"""Creates an auto-encoder wrapper."""

    def __init__(
        self,
        ae: nn.Module,  # AutoencoderDC
        scale: float = 1.0,
    ) -> None:
        super().__init__()

        self.ae = ae
        self.scale = scale

    def encode(self, x: Tensor) -> Tensor:
        r"""Encodes images to latents.

        Arguments:
            x: A batch of images :math:`x`, with shape :math:`(B, 3, H, W)`.
                Pixel values are expected to range between -1 and 1.

        Returns:
            A batch of latents :math:`z \sim E(x)`, with shape :math:`(B, 32, H / 32, W / 32)`.
        """

        dtype = get_module_dtype(self.ae)

        z = self.ae.encode(x.to(**dtype)).latent
        z = z * self.scale

        return z.to(x)

    def decode(self, z: Tensor) -> Tensor:
        r"""Decodes latents to images.

        Arguments:
            z: A batch of latents :math:`z`, with shape :math:`(B, 32, H / 32, W / 32)`.

        Returns:
            A batch of images :math:`x = D(z)`, with shape :math:`(B, 3, H, W)`.
        """

        dtype = get_module_dtype(self.ae)

        z = z / self.scale
        x = self.ae.decode(z.to(dtype)).sample

        return x.to(z)


class TextEncoder(nn.Module):
    r"""Creates a text encoder."""

    def __init__(
        self,
        gemma: nn.Module,
        tokenizer: nn.Module,
        max_length: int = 300,
    ) -> None:
        super().__init__()

        self.gemma = gemma

        self.tokenizer = tokenizer
        self.tokenizer.padding_side = "right"

        self.max_length = max_length

    def forward(
        self,
        prompt: str | Sequence[str],
        instructions: Sequence[str] = [
            "Given a user prompt, generate an 'Enhanced prompt' that provides detailed visual descriptions suitable for image generation. Evaluate the level of detail in the user prompt:",
            "- If the prompt is simple, focus on adding specifics about colors, shapes, sizes, textures, and spatial relationships to create vivid and concrete scenes.",
            "- If the prompt is already detailed, refine and enhance the existing details slightly without overcomplicating.",
            "Here are examples of how to transform or refine prompts:",
            "- User Prompt: A cat sleeping -> Enhanced: A small, fluffy white cat curled up in a round shape, sleeping peacefully on a warm sunny windowsill, surrounded by pots of blooming red flowers.",
            "- User Prompt: A busy city street -> Enhanced: A bustling city street scene at dusk, featuring glowing street lamps, a diverse crowd of people in colorful clothing, and a double-decker bus passing by towering glass skyscrapers.",
            "Please generate only the enhanced description for the prompt below and avoid including any additional commentary or evaluations:",
            "User Prompt: ",
        ],
    ) -> dict[str, Tensor]:
        r"""
        Arguments:
            prompt: A text prompt or list of text prompts.
            instructions: A set of human instructions to prepend to each prompt.

        Returns:
            The Gemma-encoded prompt(s).
        """

        if isinstance(prompt, str):
            prompt = [prompt]

        prompt = [text.lower().strip() for text in prompt]

        if instructions:
            chi = "\n".join(instructions)
            prompt = [chi + text if text else "" for text in prompt]
            max_length_all = self.max_length + len(self.tokenizer.encode(chi)) - 2
        else:
            max_length_all = self.max_length

        tokens = self.tokenizer(
            prompt,
            add_special_tokens=True,
            truncation=True,
            max_length=max_length_all,
            padding="max_length",
            return_tensors="pt",
        )

        prompt_embeds = self.gemma(
            tokens.input_ids,
            attention_mask=tokens.attention_mask,
            output_hidden_states=False,
        ).last_hidden_state
        prompt_mask = tokens.attention_mask.to(prompt_embeds)

        select = [0, *range(-self.max_length + 1, 0)]

        return {
            "prompt_embeds": prompt_embeds[:, select],
            "prompt_mask": prompt_mask[:, select],
        }


class SanaDenoiser(Denoiser):
    r"""Creates a Sana denoiser.

    Arguments:
        backbone: A time conditional network.
        schedule: A noise schedule. If :py:`None`, use
            :class:`azula.noise.DecaySchedule` instead.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = DecaySchedule()
        else:
            self.schedule = schedule

    def forward(
        self,
        z_t: Tensor,
        t: Tensor,
        prompt_embeds: Tensor,
        prompt_mask: Tensor,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            z_t: A noisy tensor :math:`z_t`, with shape :math:`(B, C, H, W)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            prompt_embeds: The Gemma-encoded text prompt :math:`y`, with shape :math:`(B, L, D)`.
            prompt_mask: The text attention mask, with shape :math:`(B, L)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Gaussian :math:`\mathcal{N}(Z \mid \mu_\phi(z_t \mid y), \Sigma_\phi(z_t \mid y)`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < z_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = 1 / (alpha_t + sigma_t)
        c_out = -sigma_t / (alpha_t + sigma_t)
        c_skip = 1 / (alpha_t + sigma_t)
        c_time = 1000 * (sigma_t / (alpha_t + sigma_t)).flatten()

        B, _, _, _ = z_t.shape
        _, L, D = prompt_embeds.shape

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            timestep=c_time.to(dtype).expand(B),
            hidden_states=(c_in * z_t).to(dtype),
            encoder_hidden_states=prompt_embeds.to(dtype).expand(B, L, D),
            encoder_attention_mask=prompt_mask.to(dtype).expand(B, L),
            **kwargs,
        ).sample.to(z_t)

        mean = c_skip * z_t + c_out * output

        return DiracPosterior(mean=mean)


def load_model(
    name: str,
    **kwargs,
) -> tuple[Denoiser, AutoEncoder, TextEncoder]:
    r"""Loads a pre-trained Sana latent denoiser.

    Arguments:
        name: The pre-trained model name.
        kwargs: Keyword arguments passed to :func:`diffusers.SanaPipeline.from_pretrained`.

    Returns:
        A pre-trained latent denoiser and the corresponding auto-encoder and text encoder.
    """

    from diffusers import SanaPipeline

    card = load_cards(__name__)[name]

    with skip_init(), patch_diffusers():
        pipe = SanaPipeline.from_pretrained(
            card.repo,
            variant=card.variant,
            torch_dtype=card.dtype_map,
            **kwargs,
        )

    patch_mmap_safetensors(pipe.transformer, pipe.vae, pipe.text_encoder)

    denoiser = SanaDenoiser(backbone=pipe.transformer)

    autoencoder = AutoEncoder(
        ae=pipe.vae,
        scale=pipe.vae.config.scaling_factor,
    )

    textencoder = TextEncoder(
        gemma=pipe.text_encoder,
        tokenizer=pipe.tokenizer,
    )

    return denoiser.eval(), autoencoder.eval(), textencoder.eval()



r"""Attention layers."""

__all__ = [
    "MultiheadSelfAttention",
]

import math
import torch

from einops import rearrange
from torch import BoolTensor, Tensor

from .layers import RMSNorm
from .utils import promote_dtype


class MultiheadSelfAttention(torch.nn.Module):
    r"""Creates a multi-head self-attention layer.

    Arguments:
        channels: The number of channels :math:`H \times C`.
        pos_channels: The number of positional channels :math:`P`.
            Only necessary with RoPE.
        attention_heads: The number of attention heads :math:`H`.
        qkv_bias: Whether to add bias to the query-key-value projection layer or not.
        qk_norm: Whether to use query-key RMS-normalization or not.
        rope: Whether to use rotary positional embedding (RoPE) or not.
        dropout: The dropout rate in :math:`[0, 1]`.
    """

    def __init__(
        self,
        channels: int,
        pos_channels: int = 1,
        attention_heads: int = 1,
        qkv_bias: bool = True,
        qk_norm: bool = True,
        rope: bool = False,
        dropout: float | None = None,
    ) -> None:
        super().__init__()

        assert channels % attention_heads == 0

        self.qkv_proj = torch.nn.Linear(channels, 3 * channels, bias=qkv_bias)
        self.y_proj = torch.nn.Linear(channels, channels, bias=False)

        if qk_norm:
            if hasattr(torch.nn, "RMSNorm"):
                self.qk_norm = torch.nn.RMSNorm(
                    channels // attention_heads,
                    elementwise_affine=False,
                    eps=1e-5,
                )
            else:
                self.qk_norm = RMSNorm(dim=-1, eps=1e-5)
        else:
            self.qk_norm = torch.nn.Identity()

        if rope:
            magnitude = torch.exp(math.log(1e-1) * torch.rand(channels // 2, 1))
            direction = torch.randn(channels // 2, pos_channels)
            direction = direction / torch.linalg.norm(direction, dim=-1, keepdim=True)

            self.theta_proj = torch.nn.Linear(pos_channels, channels // 2, bias=False)
            self.theta_proj.weight.data.copy_(magnitude * direction)
        else:
            self.theta_proj = None

        self.heads = attention_heads
        self.dropout = 0.0 if dropout is None else dropout

    def forward(
        self,
        x: Tensor,
        pos: Tensor | None = None,
        mask: BoolTensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tokens :math:`x`, with shape :math:`(*, L, H \times C)`.
            pos: Optional position vectors :math:`p`, with shape :math:`(*, L, P)`.
            mask: Optional attention mask, with shape :math:`(L, L)`.

        Returns:
            The ouput tokens :math:`y`, with shape :math:`(*, L, H \times C)`.
        """

        qkv = self.qkv_proj(x)
        q, k, v = rearrange(qkv, "... L (n H C) -> n ... H L C", n=3, H=self.heads)
        q, k = self.qk_norm(q), self.qk_norm(k)

        if self.theta_proj is not None:
            theta = self.theta_proj(pos)
            theta = rearrange(theta, "... L (H C) -> ... H L C", H=self.heads)
            q, k = apply_rope(q, k, theta)

        y = torch.nn.functional.scaled_dot_product_attention(
            query=q,
            key=k,
            value=v,
            attn_mask=mask,
            dropout_p=self.dropout if self.training else 0.0,
        )

        y = rearrange(y, "... H L C -> ... L (H C)")
        y = self.y_proj(y)

        return y


@promote_dtype
def apply_rope(q: Tensor, k: Tensor, theta: Tensor) -> tuple[Tensor, Tensor]:
    r"""
    References:
        | RoFormer: Enhanced Transformer with Rotary Position Embedding (Su et al., 2021)
        | https://arxiv.org/abs/2104.09864

        | Rotary Position Embedding for Vision Transformer (Heo et al., 2024)
        | https://arxiv.org/abs/2403.13298

    Arguments:
        q: The query vectors :math:`q`, with shape :math:`(*, C)`.
        k: The key vectors :math:`k`, with shape :math:`(*, C)`.
        theta: Rotary angles, with shape :math:`(*, C / 2)`.

    Returns:
        The rotated query and key vectors, with shape :math:`(*, C)`.
    """

    q = q.unflatten(-1, (-1, 2))
    k = k.unflatten(-1, (-1, 2))

    q_real, q_imag = q[..., 0], q[..., 1]
    k_real, k_imag = k[..., 0], k[..., 1]

    cos_theta = torch.cos(theta)
    sin_theta = torch.sin(theta)

    q = torch.stack(
        (
            q_real * cos_theta - q_imag * sin_theta,
            q_real * sin_theta + q_imag * cos_theta,
        ),
        dim=-1,
    ).flatten(-2)

    k = torch.stack(
        (
            k_real * cos_theta - k_imag * sin_theta,
            k_real * sin_theta + k_imag * cos_theta,
        ),
        dim=-1,
    ).flatten(-2)

    return q, k



import math
import torch

from functools import partial
from torch import nn
from torch.nn import functional as F


class ResidualBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return self.main(input) + self.skip(input)


class ResLinearBlock(ResidualBlock):
    def __init__(self, f_in, f_mid, f_out, is_last=False) -> None:
        skip = None if f_in == f_out else nn.Linear(f_in, f_out, bias=False)
        super().__init__(
            [
                nn.Linear(f_in, f_mid),
                nn.ReLU(inplace=True),
                nn.Linear(f_mid, f_out),
                nn.ReLU(inplace=True) if not is_last else nn.Identity(),
            ],
            skip,
        )


class Modulation2d(nn.Module):
    def __init__(self, state, feats_in, c_out) -> None:
        super().__init__()
        self.state = state
        self.layer = nn.Linear(feats_in, c_out * 2, bias=False)

    def forward(self, input):
        scales, shifts = self.layer(self.state["cond"]).chunk(2, dim=-1)
        return torch.addcmul(shifts[..., None, None], input, scales[..., None, None] + 1)


class ResModConvBlock(ResidualBlock):
    def __init__(self, state, feats_in, c_in, c_mid, c_out, is_last=False) -> None:
        skip = None if c_in == c_out else nn.Conv2d(c_in, c_out, 1, bias=False)
        super().__init__(
            [
                nn.Conv2d(c_in, c_mid, 3, padding=1),
                nn.GroupNorm(1, c_mid, affine=False),
                Modulation2d(state, feats_in, c_mid),
                nn.ReLU(inplace=True),
                nn.Conv2d(c_mid, c_out, 3, padding=1),
                nn.GroupNorm(1, c_out, affine=False) if not is_last else nn.Identity(),
                Modulation2d(state, feats_in, c_out) if not is_last else nn.Identity(),
                nn.ReLU(inplace=True) if not is_last else nn.Identity(),
            ],
            skip,
        )


class SkipBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return torch.cat([self.main(input), self.skip(input)], dim=1)


class FourierFeatures(nn.Module):
    def __init__(self, in_features, out_features, std=1.0) -> None:
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(torch.randn([out_features // 2, in_features]) * std)
        self.weight.requires_grad_(False)
        # self.register_buffer('weight', torch.randn([out_features // 2, in_features]) * std)

    def forward(self, input):
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)


class SelfAttention2d(nn.Module):
    def __init__(self, c_in, n_head=1, dropout_rate=0.1) -> None:
        super().__init__()
        assert c_in % n_head == 0
        self.norm = nn.GroupNorm(1, c_in)
        self.n_head = n_head
        self.qkv_proj = nn.Conv2d(c_in, c_in * 3, 1)
        self.out_proj = nn.Conv2d(c_in, c_in, 1)
        self.dropout = nn.Identity()  # nn.Dropout2d(dropout_rate, inplace=True)

    def forward(self, input):
        n, c, h, w = input.shape
        qkv = self.qkv_proj(self.norm(input))
        qkv = qkv.view([n, self.n_head * 3, c // self.n_head, h * w]).transpose(2, 3)
        q, k, v = qkv.chunk(3, dim=1)
        scale = k.shape[3] ** -0.25
        att = ((q * scale) @ (k.transpose(2, 3) * scale)).softmax(3)
        y = (att @ v).transpose(2, 3).contiguous().view([n, c, h, w])
        return input + self.dropout(self.out_proj(y))


def expand_to_planes(input, shape):
    return input[..., None, None].repeat([1, 1, shape[2], shape[3]])


class CC12M1Model(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.shape = (3, 256, 256)
        self.clip_model = "ViT-B/16"
        self.min_t = 0.0
        self.max_t = 1.0

        c = 128  # The base channel count
        cs = [c, c * 2, c * 2, c * 4, c * 4, c * 8, c * 8]

        self.mapping_timestep_embed = FourierFeatures(1, 128)
        self.mapping = nn.Sequential(
            ResLinearBlock(512 + 128, 1024, 1024),
            ResLinearBlock(1024, 1024, 1024, is_last=True),
        )

        with torch.no_grad():
            for param in self.mapping.parameters():
                param *= 0.5**0.5

        self.state = {}
        conv_block = partial(ResModConvBlock, self.state, 1024)

        self.timestep_embed = FourierFeatures(1, 16)
        self.down = nn.AvgPool2d(2)
        self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False)

        self.net = nn.Sequential(  # 256x256
            conv_block(3 + 16, cs[0], cs[0]),
            conv_block(cs[0], cs[0], cs[0]),
            conv_block(cs[0], cs[0], cs[0]),
            conv_block(cs[0], cs[0], cs[0]),
            SkipBlock([
                self.down,  # 128x128
                conv_block(cs[0], cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[1]),
                SkipBlock([
                    self.down,  # 64x64
                    conv_block(cs[1], cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[2]),
                    SkipBlock([
                        self.down,  # 32x32
                        conv_block(cs[2], cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[3]),
                        SkipBlock([
                            self.down,  # 16x16
                            conv_block(cs[3], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            SkipBlock([
                                self.down,  # 8x8
                                conv_block(cs[4], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                SkipBlock([
                                    self.down,  # 4x4
                                    conv_block(cs[5], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 64),
                                    conv_block(cs[6], cs[6], cs[5]),
                                    SelfAttention2d(cs[5], cs[5] // 64),
                                    self.up,
                                ]),
                                conv_block(cs[5] * 2, cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 64),
                                conv_block(cs[5], cs[5], cs[4]),
                                SelfAttention2d(cs[4], cs[4] // 64),
                                self.up,
                            ]),
                            conv_block(cs[4] * 2, cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 64),
                            conv_block(cs[4], cs[4], cs[3]),
                            SelfAttention2d(cs[3], cs[3] // 64),
                            self.up,
                        ]),
                        conv_block(cs[3] * 2, cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[3]),
                        conv_block(cs[3], cs[3], cs[2]),
                        self.up,
                    ]),
                    conv_block(cs[2] * 2, cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[2]),
                    conv_block(cs[2], cs[2], cs[1]),
                    self.up,
                ]),
                conv_block(cs[1] * 2, cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[1]),
                conv_block(cs[1], cs[1], cs[0]),
                self.up,
            ]),
            conv_block(cs[0] * 2, cs[0], cs[0]),
            conv_block(cs[0], cs[0], cs[0]),
            conv_block(cs[0], cs[0], cs[0]),
            conv_block(cs[0], cs[0], 3, is_last=True),
        )

        with torch.no_grad():
            for param in self.net.parameters():
                param *= 0.5**0.5

    def forward(self, input, t, clip_embed):
        clip_embed = F.normalize(clip_embed, dim=-1) * clip_embed.shape[-1] ** 0.5
        mapping_timestep_embed = self.mapping_timestep_embed(t[:, None])
        self.state["cond"] = self.mapping(torch.cat([clip_embed, mapping_timestep_embed], dim=1))
        timestep_embed = expand_to_planes(self.timestep_embed(t[:, None]), input.shape)
        out = self.net(torch.cat([input, timestep_embed], dim=1))
        self.state.clear()
        return out



r"""Classifier-free guidance (CFG) internals.

References:
    | Classifier-Free Diffusion Guidance (Ho et al., 2022)
    | https://arxiv.org/abs/2207.12598
"""

__all__ = [
    "CFGDenoiser",
]

from torch import Tensor
from typing import Any

from ..denoise import Denoiser, DiracPosterior
from ..noise import Schedule


class CFGDenoiser(Denoiser):
    r"""Creates a CFG denoiser module.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
    """

    def __init__(self, denoiser: Denoiser) -> None:
        super().__init__()

        self.denoiser = denoiser

    @property
    def schedule(self) -> Schedule:
        return self.denoiser.schedule

    def forward(
        self,
        x_t: Tensor,
        t: Tensor,
        positive: dict[str, Any],
        negative: dict[str, Any] = {},  # noqa: B006
        guidance: float | Tensor = 1.0,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            positive: The positive label :math:`c_+` as a dictionary of keyword arguments.
            negative: The negative label :math:`c_-` as a dictionary of keyword arguments.
            guidance: The classifier-free guidance strength :math:`\omega \in \mathbb{R}_+`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \mu)` where

            .. math:: \mu = (1 + \omega) \, \mu_\phi(x_t \mid c_+)
                - \omega \, \mu_\phi(x_t \mid c_-) \, .
        """

        q_pos = self.denoiser(x_t, t, **positive, **kwargs)
        q_neg = self.denoiser(x_t, t, **negative, **kwargs)

        return DiracPosterior(
            mean=q_pos.mean + guidance * (q_pos.mean - q_neg.mean),
        )



r"""Covariance matrices."""

from __future__ import annotations

__all__ = [
    "Covariance",
    "IsotropicCovariance",
    "DiagonalCovariance",
    "FullCovariance",
    "DPLRCovariance",
    "KroneckerCovariance",
]

import abc
import math
import torch

from collections.abc import Sequence
from torch import Tensor


class Covariance(abc.ABC):
    r"""Abstract covariance matrix."""

    @abc.abstractmethod
    def __add__(self, other: Covariance) -> Covariance:
        pass

    def __radd__(self, other: Covariance) -> Covariance:
        return self.__add__(other)

    @abc.abstractmethod
    def __mul__(self, other: Covariance) -> Covariance:
        pass

    def __rmul__(self, other: Covariance) -> Covariance:
        return self.__mul__(other)

    @abc.abstractmethod
    def __matmul__(self, x: Tensor) -> Tensor:
        pass

    def __call__(self, x: Tensor) -> Tensor:
        return self.__matmul__(x)

    @property
    @abc.abstractmethod
    def inv(self) -> Covariance:
        pass

    def to(self, **kwargs) -> Covariance:
        new = object.__new__(type(self))

        for k, v in self.__dict__.items():
            if hasattr(v, "to"):
                new.__dict__[k] = v.to(**kwargs)
            elif isinstance(v, list | tuple):
                new.__dict__[k] = type(v)(w.to(**kwargs) if hasattr(w, "to") else w for w in v)
            else:
                new.__dict__[k] = v

        return new


class IsotropicCovariance(Covariance):
    r"""Isotropic covariance matrix.

    .. math:: \lambda I
    """

    lmbda: Tensor

    def __init__(self, lmbda: Tensor) -> None:
        self.lmbda = lmbda.reshape(())

    @classmethod
    @torch.no_grad()
    def from_data(self, X: Tensor) -> IsotropicCovariance:
        return IsotropicCovariance(X.var())

    def __add__(self, other: Covariance) -> IsotropicCovariance:
        if isinstance(other, IsotropicCovariance):
            return IsotropicCovariance(self.lmbda + other.lmbda)
        else:
            return NotImplemented

    def __mul__(self, other: Covariance) -> IsotropicCovariance:
        if isinstance(other, IsotropicCovariance):
            return IsotropicCovariance(self.lmbda * other.lmbda)
        else:
            return NotImplemented

    def __matmul__(self, x: Tensor) -> Tensor:
        return self.lmbda * x

    @property
    def inv(self) -> IsotropicCovariance:
        return IsotropicCovariance(1 / self.lmbda)


class DiagonalCovariance(Covariance):
    r"""Diagonal covariance matrix.

    .. math:: \mathrm{diag}(D)
    """

    D: Tensor

    def __init__(self, D: Tensor) -> None:
        self.D = D

    @classmethod
    @torch.no_grad()
    def from_data(self, X: Tensor) -> DiagonalCovariance:
        return DiagonalCovariance(X.var(dim=0))

    def __add__(self, other: Covariance) -> DiagonalCovariance:
        if isinstance(other, IsotropicCovariance):
            return DiagonalCovariance(self.D + other.lmbda)
        if isinstance(other, DiagonalCovariance):
            return DiagonalCovariance(self.D + other.D)
        else:
            return NotImplemented

    def __mul__(self, other: Covariance) -> DiagonalCovariance:
        if isinstance(other, IsotropicCovariance):
            return DiagonalCovariance(self.D * other.lmbda)
        if isinstance(other, DiagonalCovariance):
            return DiagonalCovariance(self.D * other.D)
        else:
            return NotImplemented

    def __matmul__(self, x: Tensor) -> Tensor:
        return self.D * x

    @property
    def inv(self) -> DiagonalCovariance:
        return DiagonalCovariance(1 / self.D)


class FullCovariance(Covariance):
    r"""Full covariance matrix.

    .. math:: C
    """

    C: Tensor

    def __init__(self, C: Tensor) -> None:
        self.C = C

    @classmethod
    @torch.no_grad()
    def from_data(self, X: Tensor) -> FullCovariance:
        samples, *_ = X.shape

        C = torch.cov(X.reshape(samples, -1).T)

        return FullCovariance(C)

    def __add__(self, other: Covariance) -> FullCovariance:
        I = torch.eye(*self.C.shape, dtype=self.C.dtype, device=self.C.device)

        if isinstance(other, IsotropicCovariance):
            return FullCovariance(self.C + other.lmbda * I)
        elif isinstance(other, DiagonalCovariance):
            return FullCovariance(self.C + torch.diag_embed(other.D.flatten()))
        else:
            return NotImplemented

    def __mul__(self, other: Covariance) -> FullCovariance:
        if isinstance(other, IsotropicCovariance):
            return FullCovariance(self.C * other.lmbda)
        else:
            return NotImplemented

    def __matmul__(self, x: Tensor) -> Tensor:
        X = x.reshape(-1, self.C.shape[0])

        X = torch.einsum("ij,...j->...i", self.C, X)

        return X.reshape_as(x)

    @property
    def inv(self) -> FullCovariance:
        return FullCovariance(torch.linalg.inv(self.C))


class DPLRCovariance(Covariance):
    r"""Diagonal plus low-rank (DPLR) covariance matrix.

    .. math:: \mathrm{diag}(D) + V \Sigma V^\top

    Wikipedia:
        https://wikipedia.org/wiki/Low-rank_approximation
    """

    D: Tensor
    V: Tensor
    S: Tensor

    def __init__(self, D: Tensor, V: Tensor, S: Tensor | None = None) -> None:
        self.D, self.V = D, V

        if S is None:
            self.S = V.new_ones(self.rank)
        else:
            self.S = S

    @classmethod
    @torch.no_grad()
    def from_data(self, X: Tensor, rank: int = 1) -> DPLRCovariance:
        samples, *shape = X.shape
        features = math.prod(shape)

        X = X.flatten(1)
        X = X - X.mean(dim=0)

        if samples < features:
            C = torch.einsum("ik,jk->ij", X, X) / samples
        else:
            C = torch.einsum("ki,kj->ij", X, X) / samples

        if 3 * rank < min(samples, features):
            L, Q = torch.lobpcg(C, k=rank)
        else:
            L, Q = torch.linalg.eigh(C)
            L, Q = L[-rank:], Q[:, -rank:]

        if rank < features:
            D = (torch.trace(C) - torch.sum(L)) / (features - rank)
        else:
            D = torch.tensor(1e-6).to(C)

        if samples < features:
            V = torch.einsum("ij,ik->kj", X, Q)
            V = V / torch.linalg.norm(V, dim=1, keepdim=True)
        else:
            V = Q.T

        S = torch.clip(L - D, min=0.0)

        return DPLRCovariance(D.expand(shape), V.reshape(-1, *shape), S)

    def __add__(self, other: Covariance) -> DPLRCovariance:
        if isinstance(other, IsotropicCovariance):
            return DPLRCovariance(self.D + other.lmbda, self.V, self.S)
        elif isinstance(other, DiagonalCovariance):
            return DPLRCovariance(self.D + other.D, self.V, self.S)
        elif isinstance(other, DPLRCovariance):
            return DPLRCovariance(
                self.D + other.D,
                torch.cat((self.V, other.V)),
                torch.cat((self.S, other.S)),
            )
        else:
            return NotImplemented

    def __mul__(self, other: Covariance) -> DiagonalCovariance:
        if isinstance(other, IsotropicCovariance):
            return DPLRCovariance(self.D * other.lmbda, self.V, self.S * other.lmbda)
        else:
            return NotImplemented

    def __matmul__(self, x: Tensor) -> Tensor:
        X = x.reshape(-1, *self.D.shape)

        X = self.D * X + torch.einsum(
            "i...,i,ni->n...", self.V, self.S, torch.einsum("i...,n...->ni", self.V, X)
        )

        return X.reshape_as(x)

    @property
    def rank(self) -> int:
        return self.V.shape[0]

    @property
    def C(self) -> Tensor:  # capacitance
        return torch.diag(1 / self.S) + torch.einsum(
            "i...,...,j...->ij", self.V, 1 / self.D, self.V
        )

    @property
    def inv(self) -> DPLRCovariance:
        D = 1 / self.D
        L, Q = torch.linalg.eigh(self.C)
        V = torch.einsum("...,i...,ij->j...", D, self.V, Q)
        S = -1 / L

        return DPLRCovariance(D, V, S)


class KroneckerCovariance(Covariance):
    r"""Kronecker-factorized covariance matrix.

    .. math:: (Q_1 \otimes \dots \otimes Q_n) \, C \, (Q_1 \otimes \dots \otimes Q_n)^\top

    where :math:`Q_i` are orthonormal matrices for each dimension and :math:`\otimes` denotes the Kronecker product.

    Wikipedia:
        https://wikipedia.org/wiki/Kronecker_product
    """

    C: Covariance
    Qs: Sequence[Tensor]

    def __init__(self, C: Covariance, Qs: Sequence[Tensor]) -> None:
        self.C = C
        self.Qs = tuple(Qs)

    @classmethod
    @torch.no_grad()
    def from_data(self, X: Tensor, rank: int = 0) -> KroneckerCovariance:
        Qs = []

        for i in range(1, X.ndim):
            Ci = torch.cov(X.movedim(i, 0).flatten(1))
            _, Qi = torch.linalg.eigh(Ci)
            Qs.append(Qi)

        for Q in Qs:
            X = torch.tensordot(X, Q, dims=[[1], [0]])

        if rank > 0:
            C = DPLRCovariance.from_data(X, rank=rank)
        else:
            C = DiagonalCovariance.from_data(X)

        return KroneckerCovariance(C, Qs)

    def __add__(self, other: Covariance) -> KroneckerCovariance:
        if isinstance(other, IsotropicCovariance):
            return KroneckerCovariance(self.C + other, self.Qs)
        elif isinstance(other, KroneckerCovariance):
            assert all(Q1 is Q2 for Q1, Q2 in zip(self.Qs, other.Qs, strict=True))
            return KroneckerCovariance(
                self.C + other.C,
                self.Qs,
            )
        else:
            return NotImplemented

    def __mul__(self, other: Covariance) -> KroneckerCovariance:
        if isinstance(other, IsotropicCovariance):
            return KroneckerCovariance(self.C * other, self.Qs)
        else:
            return NotImplemented

    def __matmul__(self, x: Tensor) -> Tensor:
        X = x.reshape(-1, *(Q.shape[0] for Q in self.Qs))

        for Q in self.Qs:
            X = torch.tensordot(X, Q, dims=[[1], [0]])

        X = self.C @ X

        for Q in self.Qs:
            X = torch.tensordot(X, Q, dims=[[1], [1]])

        return X.reshape_as(x)

    @property
    def inv(self) -> KroneckerCovariance:
        return KroneckerCovariance(self.C.inv, self.Qs)



r"""Diffusion Transformer (DiT) building blocks.

References:
    | Scalable Diffusion Models with Transformers (Peebles et al., 2022)
    | https://arxiv.org/abs/2212.09748
"""

__all__ = [
    "DiT",
    "DiTBlock",
]

import torch

from einops.layers.torch import Rearrange
from torch import Tensor
from typing import Literal

from .attention import MultiheadSelfAttention
from .layers import ReLU2, RMSNorm, SineEncoding, SwiGLU
from .utils import checkpoint


class DiTBlock(torch.nn.Module):
    r"""Creates a modulated DiT block module.

    Arguments:
        channels: The number of channels :math:`C`.
        mod_features: The number of modulating features :math:`D`.
        ffn_factor: The channel factor in the FFN.
        ffn_activation: The activation function in the FFN.
        dropout: The dropout rate in :math:`[0, 1]`.
        checkpointing: Whether to use activation checkpointing or not.
        kwargs: Keyword arguments passed to :class:`MultiheadSelfAttention`.
    """

    def __init__(
        self,
        channels: int,
        mod_features: int = 0,
        ffn_factor: int = 4,
        ffn_activation: Literal["relu", "relu2", "silu", "swiglu"] = "silu",
        dropout: float | None = None,
        checkpointing: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()

        self.checkpointing = checkpointing

        # Ada-Norm Zero
        if hasattr(torch.nn, "RMSNorm"):
            self.norm = torch.nn.RMSNorm(channels, elementwise_affine=False, eps=1e-5)
        else:
            self.norm = RMSNorm(dim=-1, eps=1e-5)

        if mod_features > 0:
            self.ada_zero = torch.nn.Sequential(
                torch.nn.Linear(mod_features, mod_features),
                torch.nn.SiLU(),
                torch.nn.Linear(mod_features, 3 * channels),
                Rearrange("... (n C) -> n ... 1 C", n=3),
            )

            self.ada_zero[-2].weight.data.mul_(1e-2)
        else:
            self.ada_zero = torch.nn.Parameter(torch.randn(3, channels))
            self.ada_zero.data.mul_(1e-2)

        # MSA
        self.msa = MultiheadSelfAttention(channels, **kwargs)

        # FFN
        activation_factor = 1

        if ffn_activation == "relu":
            activation = torch.nn.ReLU()
        elif ffn_activation == "relu2":
            activation = ReLU2()
        elif ffn_activation == "silu":
            activation = torch.nn.SiLU()
        elif ffn_activation == "swiglu":
            activation = SwiGLU()
            activation_factor = 2
        else:
            raise NotImplementedError(f"Unknown activation '{ffn_activation}'.")

        self.ffn = torch.nn.Sequential(
            torch.nn.Linear(channels, ffn_factor * channels),
            activation,
            torch.nn.Identity() if dropout is None else torch.nn.Dropout(dropout),
            torch.nn.Linear(ffn_factor * channels // activation_factor, channels),
        )

    def _forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
        pos: Tensor | None = None,
        mask: Tensor | None = None,
    ) -> Tensor:
        if torch.is_tensor(self.ada_zero):
            a, b, c = self.ada_zero
        else:
            a, b, c = self.ada_zero(mod)

        y = (a + 1) * self.norm(x) + b
        y = y + self.msa(y, pos, mask)
        y = self.ffn(y)
        y = x + c * y

        return y

    def forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
        pos: Tensor | None = None,
        mask: Tensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tokens :math:`x`, with shape :math:`(*, L, C)`.
            mod: The modulation vector, with shape :math:`(D)` or :math:`(*, D)`.
            pos: The postition coordinates, with shape :math:`(*, L, N)`.
            mask: The attention mask, with shape :math:`(*, L, L)`.

        Returns:
            The ouput tokens :math:`y`, with shape :math:`(*, L, C)`.
        """
        if self.checkpointing:
            return checkpoint(self._forward, reentrant=not self.training)(x, mod, pos, mask)
        else:
            return self._forward(x, mod, pos, mask)


class DiT(torch.nn.Module):
    r"""Creates a modulated DiT-like module.

    Arguments:
        in_channels: The number of input channels :math:`C_i`.
        out_channels: The number of output channels :math:`C_o`.
        cond_channels: The number of condition channels :math:`C_c`.
        mod_features: The number of modulating features :math:`D`.
        pos_channels: The number of positional channels :math:`P`.
        hid_channels: The numbers of hidden token channels :math:`C_h`.
        hid_blocks: The number of hidden transformer blocks.
        kwargs: Keyword arguments passed to :class:`DiTBlock`.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        cond_channels: int = 0,
        mod_features: int = 0,
        pos_channels: int = 1,
        hid_channels: int = 1024,
        hid_blocks: int = 3,
        **kwargs,
    ) -> None:
        super().__init__()

        self.in_proj = torch.nn.Linear(in_channels + cond_channels, hid_channels)
        self.out_proj = torch.nn.Linear(hid_channels, out_channels)

        self.pos_embedding = torch.nn.Sequential(
            SineEncoding(hid_channels, omega=1e2),
            Rearrange("... P C -> ... (P C)"),
            torch.nn.Linear(pos_channels * hid_channels, hid_channels, bias=False),
        )
        self.pos_embedding[-1].weight.data.mul_(1e-2)

        self.blocks = torch.nn.ModuleList([
            DiTBlock(
                channels=hid_channels,
                pos_channels=pos_channels,
                mod_features=mod_features,
                **kwargs,
            )
            for _ in range(hid_blocks)
        ])

    def forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
        pos: Tensor | None = None,
        cond: Tensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tensor, with shape :math:`(*, L, C_i)`.
            mod: The modulation vector, with shape :math:`(D)` or :math:`(*, D)`.
            pos: The position tensor, with shape :math:`(*, L, P)`.
                If `None`, use the sequence indices instead.
            cond: The condition tensor, with shape :math:`(*, L, C_c)`.

        Returns:
            The output tensor, with shape :math:`(*, L, C_o)`.
        """
        if cond is not None:
            x = torch.cat((x, cond), dim=-1)

        x = self.in_proj(x)

        if pos is None:
            pos = torch.arange(x.shape[-2], dtype=x.dtype, device=x.device)
            pos = pos[..., None]

        x = x + self.pos_embedding(pos)

        for block in self.blocks:
            x = block(x, mod, pos=pos)

        x = self.out_proj(x)

        return x



r"""Reverse diffusion samplers.

For a distribution :math:`p(X)` over :math:`\mathbb{R}^D`, the perturbation kernel (see
:mod:`azula.noise`)

.. math:: p(X_t \mid X) = \mathcal{N}(X_t \mid \alpha_t X, \sigma_t^2 I)

defines a series of marginal distributions

.. math:: p(X_t) = \int p(X_t \mid x) \, p(x) \, dx \, .

The goal of diffusion models is to generate samples from :math:`p(X_0)`. To this end,
reverse transition kernels :math:`q(X_s \mid X_t)` from :math:`t` to :math:`s < t` are
chosen. Then, starting from :math:`x_1 \sim p(X_1)`, :math:`T` transitions
:math:`x_{t_{i-1}} \sim q(X_{t_{i-1}} \mid x_{t_i})` are simulated from :math:`t_T = 1`
to :math:`t_0 = 0`. If the kernels are consistent with the marginals :math:`p(X_t)`,
that is if

.. math:: p(X_{t_{i-1}}) \approx
    \int q(X_{t_{i-1}} \mid x_{t_i}) \, p(x_{t_i}) \, dx_{t_i} \, ,

for all :math:`i = 1, \dots, T`, the tensors :math:`x_{t_i}` are distributed according
to :math:`p(X_{t_i})`, including the last one :math:`x_{t_0} = x_0`.
"""

__all__ = [
    "Sampler",
    "DDPMSampler",
    "DDIMSampler",
    "EulerSampler",
    "HeunSampler",
    "ItoSampler",
    "zABSampler",
    "vABSampler",
    "zEABSampler",
    "xEABSampler",
    "REABSampler",
    "PCSampler",
]

import abc
import math
import torch

from collections.abc import Iterable, Sequence
from functools import partial
from torch import Tensor
from tqdm import tqdm

from .denoise import Denoiser
from .nn.utils import promote_dtype


class Sampler(abc.ABC):
    r"""Abstract reverse diffusion sampler.

    Arguments:
        start: The starting time :math:`t_T`.
        stop: The stopping time :math:`t_0`.
        steps: The number of discretization steps :math:`T`. By default, the step size
            :math:`t_i - t_{i-1}` is constant.
        silent: Whether to hide the sampling progress bar or not.
        dtype: The time data type.
        device: The time device.
    """

    denoiser: Denoiser

    def __init__(
        self,
        start: float = 1.0,
        stop: float = 0.0,
        steps: int = 64,
        silent: bool = False,
        dtype: torch.dtype | None = None,
        device: torch.device | None = None,
    ) -> None:
        self.start = start
        self.stop = stop
        self.steps = steps
        self.silent = silent

        self.dtype = dtype
        self.device = device

    @property
    def timesteps(self) -> Tensor:
        return torch.linspace(
            self.start,
            self.stop,
            self.steps + 1,
            dtype=self.dtype,
            device=self.device,
        )

    @torch.no_grad()
    def init(
        self,
        shape: Sequence[int],
        mean: float | Tensor = 0.0,
        var: float | Tensor = 1.0,
        **kwargs,
    ) -> Tensor:
        r"""Draws an initial noisy tensor :math:`x_{t_T}`.

        .. math:: x_{t_T} \sim \mathcal{N}(\alpha_{t_T} \mathbb{E}[X], \alpha_{t_T}^2 \mathbb{V}[X] + \sigma_{t_T}^2 I)

        Arguments:
            shape: The shape :math:`(*)` of the tensor.
            mean: The mean :math:`\mathbb{E}[X]` of :math:`p(X)`, with shape
                :math:`()` or :math:`(*)`.
            var: The variance :math:`\mathbb{V}[X]` of :math:`p(X)`, with shape
                :math:`()` or :math:`(*)`.
            kwargs: Keyword arguments passed to :func:`torch.Tensor.to`.

        Returns:
            A noisy tensor :math:`x_{t_T}`, with shape :math:`(*)`.
        """

        t_T = self.timesteps[0]

        alpha_T, sigma_T = self.denoiser.schedule(t_T)
        alpha_T, sigma_T = alpha_T.to(**kwargs), sigma_T.to(**kwargs)

        mean_T, std_T = alpha_T * mean, torch.sqrt(alpha_T**2 * var + sigma_T**2)
        mean_T, std_T = mean_T.expand(shape), std_T.expand(shape)

        return mean_T + std_T * torch.randn_like(mean_T)

    def progress_bar(self, it: Iterable) -> Iterable:
        if torch.is_tensor(it):
            it = it.unbind()

        if self.silent:
            return it
        else:
            return tqdm(it, miniters=1, unit="step", ncols=79, ascii=True)

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        r"""Simulates the reverse process from :math:`t_T` to :math:`t_0`.

        Arguments:
            x: A noisy tensor :math:`x_{t_T}`, with shape :math:`(*)`.
            kwargs: Optional keyword arguments.

        Returns:
            The clean(er) tensor :math:`x_{t_0}`, with shape :math:`(*)`.
        """

        time_pairs = self.timesteps.unfold(0, 2, 1).to(device=x.device)

        x_t = x

        for t, s in self.progress_bar(time_pairs):
            x_s = self.step(x_t, t, s, **kwargs)
            x_t = x_s

        x = x_t

        return x

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        r"""Simulates the reverse process from :math:`t` to :math:`s`.

        Arguments:
            x_t: The current tensor :math:`x_t`, with shape :math:`(*)`.
            t: The current time :math:`t`, with shape :math:`()`.
            s: The target time :math:`s`, with shape :math:`()`.
            kwargs: Optional keyword arguments.

        Returns:
            The new tensor :math:`x_s \sim q(X_s \mid x_t)`, with shape :math:`(*)`.
        """

        raise NotImplementedError()


class DDPMSampler(Sampler):
    r"""Creates an DDPM sampler.

    .. math:: x_s \gets \alpha_s \mathbb{E}[X \mid x_t]
        + \sigma_s \, \sqrt{1 - \tau} \, \frac{x_t - \alpha_t \mathbb{E}[X \mid x_t]}{\sigma_t}
        + \sigma_s \, \sqrt{\tau} \, \varepsilon

    where :math:`\varepsilon \sim \mathcal{N}(0, I)` and

    .. math:: \tau = 1 - \frac{\alpha_t^2}{\alpha_s^2} \frac{\sigma_s^2}{\sigma_t^2} \, .

    References:
        | Denoising Diffusion Probabilistic Models (Ho et al., 2020)
        | https://arxiv.org/abs/2006.11239

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        tau = 1 - (alpha_t / alpha_s * sigma_s / sigma_t) ** 2

        q_t = self.denoiser(x_t, t, **kwargs)

        x_s = alpha_s * q_t.mean
        x_s = x_s + sigma_s * torch.sqrt(1 - tau) / sigma_t * (x_t - alpha_t * q_t.mean)
        x_s = x_s + sigma_s * torch.sqrt(tau) * torch.randn_like(x_t)

        return x_s


class DDIMSampler(Sampler):
    r"""Creates a DDIM sampler.

    .. math:: x_s \gets \alpha_s \mathbb{E}[X \mid x_t]
        + \sigma_s \, \sqrt{1 - \eta \, \tau} \, \frac{x_t - \alpha_t \mathbb{E}[X \mid x_t]}{\sigma_t}
        + \sigma_s \, \sqrt{\eta \, \tau} \, \varepsilon

    where :math:`\varepsilon \sim \mathcal{N}(0, I)` and

    .. math:: \tau = 1 - \frac{\alpha_t^2}{\alpha_s^2} \frac{\sigma_s^2}{\sigma_t^2} \, .

    References:
        | Denoising Diffusion Implicit Models (Song et al., 2021)
        | https://arxiv.org/abs/2010.02502

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        eta: The stochasticity parameter :math:`\eta \in \mathbb{R}_+`.
            If :math:`\eta = 1`, :class:`DDIMSampler` is equivalent to :class:`DDPMSampler`.
            If :math:`\eta = 0`, :class:`DDIMSampler` is equivalent to :class:`EulerSampler`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, eta: float = 0.0, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.eta = eta

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        tau = 1 - (alpha_t / alpha_s * sigma_s / sigma_t) ** 2
        tau = torch.clip(self.eta * tau, min=0, max=1)

        q_t = self.denoiser(x_t, t, **kwargs)

        x_s = alpha_s * q_t.mean
        x_s = x_s + sigma_s * torch.sqrt(1 - tau) / sigma_t * (x_t - alpha_t * q_t.mean)
        x_s = x_s + sigma_s * torch.sqrt(tau) * torch.randn_like(x_t)

        return x_s


class EulerSampler(Sampler):
    r"""Creates a explicit Euler (1st order) sampler.

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = \mathbb{E}[Z \mid x_t] = z(x_t)

    and

    .. math:: x_s = x_t + \int_t^s z(x_u) \, du \, .

    The explicit Euler step for this integral is

    .. math:: x_s \gets x_t + (s - t) \, z(x_t)

    Wikipedia:
        https://wikipedia.org/wiki/Euler_method

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        q_t = self.denoiser(x_t, t, **kwargs)
        z_t = (x_t - alpha_t * q_t.mean) / sigma_t
        x_s = alpha_s / alpha_t * x_t + alpha_s * (sigma_s / alpha_s - sigma_t / alpha_t) * z_t

        return x_s


class HeunSampler(Sampler):
    r"""Creates a explicit Heun (2nd order) sampler.

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = \mathbb{E}[Z \mid x_t] = z(x_t)

    and

    .. math:: x_s = x_t + \int_t^s z(x_u) \, du \, .

    The explicit Heun step for this integral is

    .. math::
        x_s & \gets x_t + (s - t) \, z(x_t) \\
        x_s & \gets x_t + (s - t) \frac{z(x_t) + z(x_s)}{2}

    Wikipedia:
        https://wikipedia.org/wiki/Heun%27s_method

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        q_t = self.denoiser(x_t, t, **kwargs)
        z_t = (x_t - alpha_t * q_t.mean) / sigma_t
        x_s = alpha_s / alpha_t * x_t + alpha_s * (sigma_s / alpha_s - sigma_t / alpha_t) * z_t

        q_s = self.denoiser(x_s, s, **kwargs)
        z_s = (x_s - alpha_s * q_s.mean) / sigma_s
        z_t = (z_t + z_s) / 2
        x_s = alpha_s / alpha_t * x_t + alpha_s * (sigma_s / alpha_s - sigma_t / alpha_t) * z_t

        return x_s


class ItoSampler(Sampler):
    r"""Creates an Itô SDE sampler.

    Let's consider the Itô SDE

    .. math:: dx_t = \left[ f_t \, x_t - \frac{1 + \eta^2}{2 \tau} g_t^2 \, \nabla_{x_t} \log p(x_t) \right] dt + \eta \, g_t \, dw_t

    where :math:`\eta \geq 0` controls stochasticity, :math:`\tau \geq 0` controls
    temperature, and

    .. math::
        f_t & = \partial_t \log \alpha_t \\
        g_t^2 & = \alpha_t^2 \, \partial_t \frac{\sigma_t^2}{\alpha_t^2} \, .

    The integral form of this semi-linear Itô SDE is

    .. math:: x_s = \Psi(t, s) \, x_t
        + \frac{1 + \eta^2}{2} \int_t^s \Psi(u, s) \, g_u^2 \, \nabla_{x_u} \log p(x_u) \, du
        + \eta \int_t^s \Psi(u, s) \, g_u \, dw_u

    where

    .. math:: \Psi(t, s) = \exp \left( \int_t^s f_u \, du \right) = \frac{\alpha_s}{\alpha_t} \, .

    By Tweedie's formula, we have

    .. math:: \int_t^s \Psi(u, s) \, g_u^2 \, \nabla_{x_u} \log p(x_u) \, du
        & = \int_t^s \Psi(u, s) \, g_u^2 \frac{\alpha_u \mathbb{E}[x_0 | x_u] - x_u}{\sigma_u^2} du \\
        & = 2 \alpha_s \int_t^s \partial_u \frac{\sigma_u}{\alpha_u} \frac{\alpha_u \mathbb{E}[x_0 | x_u] - x_u}{\sigma_u} du \\
        & \approx 2 \alpha_s \frac{\alpha_t \mathbb{E}[x_0 | x_t] - x_t}{\sigma_t} \int_t^s \partial_u \frac{\sigma_u}{\alpha_u} du \\
        & \approx 2 \left( \frac{\sigma_s}{\sigma_t} - \frac{\alpha_s}{\alpha_t} \right) (\alpha_t \mathbb{E}[x_0 | x_t] - x_t) \, .

    Finally, by Itô isometry, we have

    .. math:: \int_t^s \Psi(u, s) \, g_u \, dw_u
        & = \int_s^t \Psi(u, s) \, g_u \, dw_u \\
        & = \varepsilon \, \sqrt{\int_s^t \Psi(u, s)^2 \, g_u^2 \, du} \\
        & = \varepsilon \, \sqrt{\int_s^t \alpha_s^2 \, \partial_u \frac{\sigma_u^2}{\alpha_u^2} du} \\
        & = \varepsilon \, \alpha_s \sqrt{\frac{\sigma_t^2}{\alpha_t^2} - \frac{\sigma_s^2}{\alpha_s^2}}

    where :math:`\varepsilon \sim \mathcal{N}(0, 1)`.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        eta: The stochasticity parameter :math:`\eta \geq 0`.
        temperature: The temperature parameter :math:`\tau \geq 0`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        eta: float = 1.0,
        temperature: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.eta = eta
        self.temperature = temperature

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        q_t = self.denoiser(x_t, t, **kwargs)

        x_s = alpha_s / alpha_t * x_t
        x_s = x_s + (1 + self.eta**2) / self.temperature * (
            sigma_s / sigma_t - alpha_s / alpha_t
        ) * (x_t - alpha_t * q_t.mean)
        x_s = x_s + self.eta * alpha_s * torch.sqrt(
            torch.abs((sigma_t / alpha_t) ** 2 - (sigma_s / alpha_s) ** 2)
        ) * torch.randn_like(x_s)

        return x_s


class zABSampler(Sampler):
    r"""Creates an Adams-Bashforth (AB) multi-step sampler with noise (:math:`z`) prediction.

    Note:
        This sampler is equivalent to the :math:`\rho\mathrm{AB}` sampler from Zhang et al.
        (2023) and the linear multi-step (LMS) sampler from Katherine Crowson's
        `k-diffusion <https://github.com/crowsonkb/k-diffusion>`_.

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = \mathbb{E}[Z \mid x_t] = z(x_t)

    and

    .. math:: x_s = x_t + \int_t^s z(x_u) \, du \, .

    The :math:`n`-th order Adams-Bashforth step for this integral is

    .. math:: x_s \gets x_t + \sum_{i=1}^{n} z(x_{t_i}) \int_t^s \ell_i(u) \, du

    where :math:`t_i` are previous time steps and the polynomials :math:`\ell_i(u) =
    \sum_{k=1}^{n} a_{ik} \, u^k` form their Lagrange basis. The coefficients
    :math:`a_{ik}` are determined by solving the system of linear equations
    :math:`\ell_i(t_j) = \delta_{ij}`. Then, the Adams-Bashforth coefficients are

    .. math:: \int_t^s \ell_i(u) \, du
        & = \sum_{k=1}^{n} a_{ik} \int_t^s u^k \, du \\
        & = \sum_{k=1}^{n} a_{ik} \left[ \frac{u^{k+1}}{k+1} \right]_t^s

    Wikipedia:
        https://wikipedia.org/wiki/Linear_multistep_method

    References:
        | Fast Sampling of Diffusion Models with Exponential Integrator (Zhang et al., 2023)
        | https://arxiv.org/abs/2204.13902

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        order: The order :math:`n` of the multi-step method.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, order: int = 2, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.order = order

    @staticmethod
    @partial(promote_dtype, min_dtype=torch.float64)
    def _adams_bashforth(t: Tensor, /, i: int, n: int) -> Tensor:
        r"""Returns the coefficients of the :math:`n`-th order Adams-Bashforth method.

        Arguments:
            t: The integration variable, with shape :math:`(T)`.
            i: The current step :math:`i`.
            n: The method order :math:`n \leq i + 1`.

        Returns:
            The coefficients, with shape :math:`(n)`.
        """

        n = min(n, i + 1)
        k = torch.arange(n, device=t.device)

        # Vandermonde matrix t_i^k
        V = t[i + 1 - n : i + 1] ** k[:, None]

        # Integral of u^k from t_i to t_{i+1}
        b = t[i + 1] ** (k + 1) / (k + 1) - t[i] ** (k + 1) / (k + 1)

        return torch.linalg.solve(V, b)

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        time = self.timesteps.to(device=x.device)
        alpha, sigma = self.denoiser.schedule(time)
        u = sigma / alpha

        x_t = x

        buffer = []

        for i, t in enumerate(self.progress_bar(time[:-1])):
            alpha_t, sigma_t = alpha[i], sigma[i]
            alpha_s = alpha[i + 1]

            q_t = self.denoiser(x_t, t, **kwargs)
            z_t = (x_t - alpha_t * q_t.mean) / sigma_t

            buffer.append(z_t)
            if len(buffer) > self.order:
                buffer.pop(0)

            coeffs = self._adams_bashforth(u, i=i, n=self.order)
            integral = sum(b * c for b, c in zip(buffer, coeffs, strict=True))

            x_s = alpha_s / alpha_t * x_t + alpha_s * integral
            x_t = x_s

        x = x_t

        return x


class vABSampler(zABSampler):
    r"""Creates an Adams-Bashforth (AB) multi-step sampler with velocity (:math:`v`) prediction.

    Without loss of generality, let's assume :math:`\alpha_t = 1 - t` and
    :math:`\sigma_t = t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = -\mathbb{E}[X \mid x_t] + \mathbb{E}[Z \mid x_t] = v(x_t)

    and

    .. math:: x_s = x_t + \int_t^s v(x_u) \, du \, .

    The :math:`n`-th order Adams-Bashforth step for this integral is

    .. math:: x_s \gets x_t + \sum_{i=1}^{n} v(x_{t_i}) \int_t^s \ell_i(u) \, du

    where :math:`t_i` are previous time steps and the polynomials :math:`\ell_i(u)` form their Lagrange basis.

    See also:
        :class:`zABSampler`

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        order: The order :math:`n` of the multi-step method.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        time = self.timesteps.to(device=x.device)
        alpha, sigma = self.denoiser.schedule(time)
        u = sigma / (alpha + sigma)

        x_t = x

        buffer = []

        for i, t in enumerate(self.progress_bar(time[:-1])):
            alpha_t, sigma_t = alpha[i], sigma[i]
            alpha_s, sigma_s = alpha[i + 1], sigma[i + 1]

            q_t = self.denoiser(x_t, t, **kwargs)
            v_t = 1 / sigma_t * x_t - (1 + alpha_t / sigma_t) * q_t.mean

            buffer.append(v_t)
            if len(buffer) > self.order:
                buffer.pop(0)

            coeffs = self._adams_bashforth(u, i=i, n=self.order)
            integral = sum(b * c for b, c in zip(buffer, coeffs, strict=True))

            x_s = (alpha_s + sigma_s) / (alpha_t + sigma_t) * x_t + (alpha_s + sigma_s) * integral
            x_t = x_s

        x = x_t

        return x


class zEABSampler(Sampler):
    r"""Creates an exponential Adams-Bashforth (EAB) multi-step sampler with noise (:math:`z`) prediction.

    Note:
        This sampler is a multi-step generalization of the DPM-Solver sampler from Lu et
        al. (2022).

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    e^t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = e^t \, \mathbb{E}[Z \mid x_t] = e^t \, z(x_t)

    and

    .. math:: x_s = x_t + \int_t^s e^u \, z(x_u) \, du \, .

    The :math:`n`-th order exponential Adams-Bashforth step for this integral is

    .. math:: x_s \gets x_t + \sum_{i=1}^{n} z(x_{t_i}) \int_t^s e^u \, \ell_i(u) \, du

    where :math:`t_i` are previous time steps and the polynomials :math:`\ell_i(u) =
    \sum_{k=1}^{n} a_{ik} \, u^k` form their Lagrange basis. The coefficients
    :math:`a_{ik}` are determined by solving the system of linear equations
    :math:`\ell_i(t_j) = \delta_{ij}`. Then, the exponential Adams-Bashforth
    coefficients are

    .. math:: \int_t^s e^u \, \ell_i(u) \, du
        & = \sum_{k=1}^{n} a_{ik} \int_t^s e^u \, u^k \, du \\
        & = \sum_{k=1}^{n} a_{ik} \left[ (-1)^k \, k! \, e^u \sum_{j=0}^k \frac{(-u)^j}{j!} \right]_t^s

    References:
        | DPM-Solver: A Fast ODE Solver for Diffusion Probabilistic Model Sampling in Around 10 Steps (Lu et al., 2022)
        | https://arxiv.org/abs/2206.00927

        | Exponential Adams Bashforth ODE solver for stiff problems (Coudière et al., 2018)
        | https://arxiv.org/abs/1804.09927

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        order: The order :math:`n` of the multi-step method.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, order: int = 2, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.order = order

    @staticmethod
    @partial(promote_dtype, min_dtype=torch.float64)
    def _exponential_adams_bashforth(t: Tensor, /, i: int, n: int) -> Tensor:
        r"""Returns the coefficients of the :math:`n`-th order exponential Adams-Bashforth method.

        Arguments:
            t: The integration variable, with shape :math:`(T)`.
            i: The current step :math:`i`.
            n: The method order :math:`n \leq i + 1`.

        Returns:
            The coefficients, with shape :math:`(n)`.
        """

        n = min(n, i + 1)
        k = torch.arange(n, device=t.device)
        k_fact = torch.cumprod(torch.clip(k, min=1), dim=0)

        # Vandermonde matrix t_i^k
        V = t[i + 1 - n : i + 1] ** k[:, None]

        # Integral of exp(u) u^k from t_i to t_{i+1}
        b = (
            (-1) ** k
            * k_fact
            * (
                torch.exp(t[i + 1]) * torch.cumsum((-t[i + 1]) ** k / k_fact, dim=0)
                - torch.exp(t[i]) * torch.cumsum((-t[i]) ** k / k_fact, dim=0)
            )
        )

        return torch.linalg.solve(V, b)

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        time = self.timesteps.to(device=x.device)
        alpha, sigma = self.denoiser.schedule(time)
        u = sigma.log() - alpha.log()

        x_t = x

        buffer = []

        for i, t in enumerate(self.progress_bar(time[:-1])):
            alpha_t, sigma_t = alpha[i], sigma[i]
            alpha_s = alpha[i + 1]

            q_t = self.denoiser(x_t, t, **kwargs)
            z_t = (x_t - alpha_t * q_t.mean) / sigma_t

            buffer.append(z_t)
            if len(buffer) > self.order:
                buffer.pop(0)

            coeffs = self._exponential_adams_bashforth(u, i=i, n=self.order)
            integral = sum(b * c for b, c in zip(buffer, coeffs, strict=True))

            x_s = alpha_s / alpha_t * x_t + alpha_s * integral
            x_t = x_s

        x = x_t

        return x


class xEABSampler(Sampler):
    r"""Creates an exponential Adams-Bashforth (EAB) multi-step sampler with data (:math:`x`) prediction.

    Note:
        This sampler is a multi-step generalization of the DPM-Solver++ sampler from Lu et al. (2022).

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    e^t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = e^t \, \mathbb{E}[Z \mid x_t] = x_t - \mathbb{E}[X \mid x_t] = x_t - x(x_t)

    and

    .. math:: x_s = e^{s-t} \, x_t - \int_t^s e^{s-u} \, x(x_u) \, du \, .

    The :math:`n`-th order exponential Adams-Bashforth step for this integral is

    .. math:: x_s \gets e^{s-t} \, x_t
        - e^s \sum_{i=1}^{n} x(x_{t_i}) \int_t^s e^{-u} \, \ell_i(u) \, du

    where :math:`t_i` are previous time steps and the polynomials :math:`\ell_i(u) =
    \sum_{k=1}^{n} a_{ik} \, u^k` form their Lagrange basis. The coefficients
    :math:`a_{ik}` are determined by solving the system of linear equations
    :math:`\ell_i(t_j) = \delta_{ij}`. Then, the exponential Adams-Bashforth
    coefficients are

    .. math:: \int_t^s e^{-u} \, \ell_i(u) \, du
        & = \sum_{k=1}^{n} a_{ik} \int_t^s e^{-u} \, u^k \, du \\
        & = \sum_{k=1}^{n} a_{ik} \left[ -k! \, e^{-u} \sum_{j=0}^k \frac{u^j}{j!} \right]_t^s

    References:
        | DPM-Solver++: Fast Solver for Guided Sampling of Diffusion Probabilistic Models (Lu et al., 2022)
        | https://arxiv.org/abs/2211.01095

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        order: The order :math:`n` of the multi-step method.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, order: int = 2, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.order = order

    @staticmethod
    @partial(promote_dtype, min_dtype=torch.float64)
    def _exponential_adams_bashforth(t: Tensor, /, i: int, n: int) -> Tensor:
        r"""Returns the coefficients of the :math:`n`-th order exponential Adams-Bashforth method.

        Arguments:
            t: The integration variable, with shape :math:`(T)`.
            i: The current step :math:`i`.
            n: The method order :math:`n \leq i + 1`.

        Returns:
            The coefficients, with shape :math:`(n)`.
        """

        n = min(n, i + 1)
        k = torch.arange(n, device=t.device)
        k_fact = torch.cumprod(torch.clip(k, min=1), dim=0)

        # Vandermonde matrix t_i^k
        V = t[i + 1 - n : i + 1] ** k[:, None]

        # Integral of exp(-u) u^k from t_i to t_{i+1}
        b = -k_fact * (
            torch.exp(-t[i + 1]) * torch.cumsum(t[i + 1] ** k / k_fact, dim=0)
            - torch.exp(-t[i]) * torch.cumsum(t[i] ** k / k_fact, dim=0)
        )

        return torch.linalg.solve(V, b)

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        time = self.timesteps.to(device=x.device)
        alpha, sigma = self.denoiser.schedule(time)
        u = sigma.log() - alpha.log()

        x_t = x

        buffer = []

        for i, t in enumerate(self.progress_bar(time[:-1])):
            sigma_t = sigma[i]
            sigma_s = sigma[i + 1]

            q_t = self.denoiser(x_t, t, **kwargs)

            buffer.append(q_t.mean)
            if len(buffer) > self.order:
                buffer.pop(0)

            coeffs = self._exponential_adams_bashforth(u, i=i, n=self.order)
            integral = sum(b * c for b, c in zip(buffer, coeffs, strict=True))

            x_s = sigma_s / sigma_t * x_t - sigma_s * integral
            x_t = x_s

        x = x_t

        return x


class REABSampler(Sampler):
    r"""Creates a Rosenbrock-type exponential Adams-Bashforth (REAB) multi-step sampler.

    Note:
        This sampler is a multi-step generalization of the DPM-Solver-v3 sampler from Zheng et al. (2023).

    Without loss of generality, let's assume :math:`\alpha_t = 1` and :math:`\sigma_t =
    e^t` such that

    .. math:: \frac{d x_t}{dt}
        & = \alpha_t' \, \mathbb{E}[X \mid x_t] + \sigma_t' \, \mathbb{E}[Z \mid x_t] \\
        & = e^t \, \mathbb{E}[Z \mid x_t] = x_t - \mathbb{E}[X \mid x_t] \\
        & = a_t \, x_t + b_t \frac{(1 - a_t) \, x_t - \mathbb{E}[X \mid x_t]}{b_t}
            = a_t \, x_t + b_t \, f(x_t)

    and

    .. math:: x_s = \Psi(t, s) \, x_t + \int_t^s \Psi(u, s) \, b_u \, f(x_u) \, du

    where :math:`a_t = \frac{e^{2t}}{1 + e^{2t}}`, :math:`b_t = \sqrt{a_t}` and

    .. math:: \Psi(t, s)
        = \exp \left( \int_t^s a_u \, du \right)
        = \sqrt{\frac{1 + e^{2s}}{1 + e^{2t}}} \, .

    The :math:`n`-th order exponential Adams-Bashforth step for this integral is

    .. math:: x_s \gets \sqrt{\frac{1 + e^{2s}}{1 + e^{2t}}} \, x_t
        + \sqrt{1 + e^{2s}} \sum_{i=1}^{n} f(x_{t_i}) \int_t^s \frac{e^u}{1 + e^{2u}} \ell_i(u) \, du

    where :math:`t_i` are previous time steps and the polynomials :math:`\ell_i(u) =
    \sum_{k=1}^{n} a_{ik} \, u^k` form their Lagrange basis. The coefficients
    :math:`a_{ik}` are determined by solving the system of linear equations
    :math:`\ell_i(t_j) = \delta_{ij}`. Then, the exponential Adams-Bashforth
    coefficients are

    .. math:: \int_t^s \frac{e^u}{1 + e^{2u}} \ell_i(u) \, du
        = \sum_{k=1}^{n} a_{ik} \int_t^s \frac{e^u \, u^k}{1 + e^{2u}} du

    where the last integral is estimated by numerical integration.

    References:
        | DPM-Solver-v3: Improved Diffusion ODE Solver with Empirical Model Statistics (Zheng et al., 2023)
        | https://arxiv.org/abs/2310.13268

        | Exponential Rosenbrock-Type Methods (Hochbruck et al., 2009)
        | https://epubs.siam.org/doi/10.1137/080717717

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        order: The order :math:`n` of the multi-step method.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(self, denoiser: Denoiser, order: int = 2, **kwargs) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.order = order

    @staticmethod
    @partial(promote_dtype, min_dtype=torch.float64)
    def _exponential_adams_bashforth(t: Tensor, /, i: int, n: int) -> Tensor:
        r"""Returns the coefficients of the :math:`n`-th order exponential Adams-Bashforth method.

        Arguments:
            t: The integration variable, with shape :math:`(T)`.
            i: The current step :math:`i`.
            n: The method order :math:`n \leq i + 1`.

        Returns:
            The coefficients, with shape :math:`(n)`.
        """

        n = min(n, i + 1)
        k = torch.arange(n, device=t.device)

        # Vandermonde matrix t_i^k
        V = t[i + 1 - n : i + 1] ** k[:, None]

        # Integral of exp(u) / (1 + exp(2u)) u^k from t_i to t_{i+1}
        u = torch.linspace(t[i], t[i + 1], steps=256 + 1, dtype=t.dtype, device=t.device)
        y = torch.exp(u) / (1 + torch.exp(2 * u)) * (u ** k[:, None])
        b = torch.trapezoid(y, u, dim=-1)

        return torch.linalg.solve(V, b)

    @torch.no_grad()
    def __call__(self, x: Tensor, **kwargs) -> Tensor:
        time = self.timesteps.to(device=x.device)
        alpha, sigma = self.denoiser.schedule(time)
        u = sigma.log() - alpha.log()

        x_t = x

        buffer = []

        for i, t in enumerate(self.progress_bar(time[:-1])):
            alpha_t, sigma_t = alpha[i], sigma[i]
            alpha_s, sigma_s = alpha[i + 1], sigma[i + 1]

            q_t = self.denoiser(x_t, t, **kwargs)

            a_t = sigma_t**2 / (alpha_t**2 + sigma_t**2)
            b_t = sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)

            f_t = (1 - a_t) / b_t / alpha_t * x_t - 1 / b_t * q_t.mean

            buffer.append(f_t)
            if len(buffer) > self.order:
                buffer.pop(0)

            coeffs = self._exponential_adams_bashforth(u, i=i, n=self.order)
            integral = sum(b * c for b, c in zip(buffer, coeffs, strict=True))

            x_s = (
                torch.sqrt((alpha_s**2 + sigma_s**2) / (alpha_t**2 + sigma_t**2)) * x_t
                + torch.sqrt(alpha_s**2 + sigma_t**2) * integral
            )
            x_t = x_s

        x = x_t

        return x


class PCSampler(Sampler):
    r"""Creates a predictor-corrector (PC) sampler.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        corrections: The number of corrector steps for each predictor step.
        delta: The amplitude of corrector steps :math:`\delta \in [0,1]`.
        kwargs: Keyword arguments passed to :class:`Sampler`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        corrections: int = 1,
        delta: float = 0.01,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.denoiser = denoiser
        self.corrections = corrections
        self.delta = delta

    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        # Corrector
        for _ in range(self.corrections):
            q_t = self.denoiser(x_t, t, **kwargs)
            x_t = (
                alpha_t * q_t.mean
                + math.sqrt(1 - self.delta) * (x_t - alpha_t * q_t.mean)
                + math.sqrt(self.delta) * sigma_t * torch.randn_like(x_t)
            )

        # Predictor
        q_t = self.denoiser(x_t, t, **kwargs)
        x_s = alpha_s * q_t.mean + sigma_s / sigma_t * (x_t - alpha_t * q_t.mean)

        return x_s



r"""Tests for the azula.hub module."""

import os
import pytest

from pathlib import Path

from azula.hub import download, get_hub_dir, set_hub_dir


def test_default_dir() -> None:
    default_dir = get_hub_dir()

    assert isinstance(default_dir, str)

    os.makedirs(default_dir, exist_ok=True)


def test_set_dir(tmp_path: Path) -> None:
    set_hub_dir(tmp_path)
    cache_dir = get_hub_dir()

    assert isinstance(cache_dir, str)
    assert os.path.samefile(cache_dir, tmp_path)


def test_download(tmp_path: Path) -> None:
    # Set cache dir
    set_hub_dir(tmp_path)

    # With filename
    download(
        url="https://raw.githubusercontent.com/probabilists/azula/master/LICENSE",
        filename=tmp_path / "LICENSE",
    )

    with open(tmp_path / "LICENSE") as f:
        text = f.read()

        assert "MIT License" in text
        assert "The Probabilists" in text

    # Without filename
    filename = download(
        url="https://raw.githubusercontent.com/probabilists/azula/master/LICENSE",
    )

    assert os.path.samefile(os.path.dirname(filename), tmp_path)

    with open(filename) as f:
        text = f.read()

        assert "MIT License" in text
        assert "The Probabilists" in text

    # Hash prefix
    download(
        url="https://raw.githubusercontent.com/probabilists/azula/master/LICENSE",
        hash_prefix="sha256:c8adb00fadb8f4bf",
    )

    with pytest.raises(AssertionError):
        download(
            url="https://raw.githubusercontent.com/probabilists/azula/master/LICENSE",
            hash_prefix="sha256:abcdefghijklmnop",
        )

    # TODO (francois-rozet)
    # Find a URL to test Google Drive download



# Adapted from
# https://github.com/LTH14/JiT/blob/869190a33f59271724cd1bfddd85777501dadf03/util/model_util.py

import math
import numpy as np
import torch
import torch.nn as nn

from einops import rearrange, repeat


def broadcat(tensors, dim=-1):
    num_tensors = len(tensors)
    shape_lens = set(list(map(lambda t: len(t.shape), tensors)))
    assert len(shape_lens) == 1, "tensors must all have the same number of dimensions"
    shape_len = list(shape_lens)[0]
    dim = (dim + shape_len) if dim < 0 else dim
    dims = list(zip(*map(lambda t: list(t.shape), tensors), strict=True))
    expandable_dims = [(i, val) for i, val in enumerate(dims) if i != dim]
    assert all([*map(lambda t: len(set(t[1])) <= 2, expandable_dims)]), (
        "invalid dimensions for broadcastable concatentation"
    )
    max_dims = list(map(lambda t: (t[0], max(t[1])), expandable_dims))
    expanded_dims = list(map(lambda t: (t[0], (t[1],) * num_tensors), max_dims))
    expanded_dims.insert(dim, (dim, dims[dim]))
    expandable_shapes = list(zip(*map(lambda t: t[1], expanded_dims), strict=True))
    tensors = list(map(lambda t: t[0].expand(*t[1]), zip(tensors, expandable_shapes, strict=True)))
    return torch.cat(tensors, dim=dim)


def rotate_half(x):
    x = rearrange(x, "... (d r) -> ... d r", r=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, "... d r -> ... (d r)")


class VisionRotaryEmbedding(nn.Module):
    def __init__(
        self,
        dim,
        pt_seq_len,
        ft_seq_len=None,
        custom_freqs=None,
        freqs_for="lang",
        theta=10000,
        max_freq=10,
        num_freqs=1,
    ) -> None:
        super().__init__()
        if custom_freqs:
            freqs = custom_freqs
        elif freqs_for == "lang":
            freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
        elif freqs_for == "pixel":
            freqs = torch.linspace(1.0, max_freq / 2, dim // 2) * math.pi
        elif freqs_for == "constant":
            freqs = torch.ones(num_freqs).float()
        else:
            raise ValueError(f"unknown modality {freqs_for}")

        if ft_seq_len is None:
            ft_seq_len = pt_seq_len
        t = torch.arange(ft_seq_len) / ft_seq_len * pt_seq_len

        freqs_h = torch.einsum("..., f -> ... f", t, freqs)
        freqs_h = repeat(freqs_h, "... n -> ... (n r)", r=2)

        freqs_w = torch.einsum("..., f -> ... f", t, freqs)
        freqs_w = repeat(freqs_w, "... n -> ... (n r)", r=2)

        freqs = broadcat((freqs_h[:, None, :], freqs_w[None, :, :]), dim=-1)

        self.register_buffer("freqs_cos", freqs.cos())
        self.register_buffer("freqs_sin", freqs.sin())

    def forward(self, t, start_index=0):
        rot_dim = self.freqs_cos.shape[-1]
        end_index = start_index + rot_dim
        assert rot_dim <= t.shape[-1], (
            f"feature dimension {t.shape[-1]} is not of sufficient size to rotate in all the positions {rot_dim}"
        )
        t_left, t, t_right = (
            t[..., :start_index],
            t[..., start_index:end_index],
            t[..., end_index:],
        )
        t = (t * self.freqs_cos) + (rotate_half(t) * self.freqs_sin)
        return torch.cat((t_left, t, t_right), dim=-1)


class VisionRotaryEmbeddingFast(nn.Module):
    def __init__(
        self,
        dim,
        pt_seq_len=16,
        ft_seq_len=None,
        custom_freqs=None,
        freqs_for="lang",
        theta=10000,
        max_freq=10,
        num_freqs=1,
        num_cls_token=0,
    ) -> None:
        super().__init__()
        if custom_freqs:
            freqs = custom_freqs
        elif freqs_for == "lang":
            freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
        elif freqs_for == "pixel":
            freqs = torch.linspace(1.0, max_freq / 2, dim // 2) * math.pi
        elif freqs_for == "constant":
            freqs = torch.ones(num_freqs).float()
        else:
            raise ValueError(f"unknown modality {freqs_for}")

        if ft_seq_len is None:
            ft_seq_len = pt_seq_len
        t = torch.arange(ft_seq_len) / ft_seq_len * pt_seq_len

        freqs = torch.einsum("..., f -> ... f", t, freqs)
        freqs = repeat(freqs, "... n -> ... (n r)", r=2)
        freqs = broadcat((freqs[:, None, :], freqs[None, :, :]), dim=-1)

        if num_cls_token > 0:
            freqs_flat = freqs.view(-1, freqs.shape[-1])  # [N_img, D]
            cos_img = freqs_flat.cos()
            sin_img = freqs_flat.sin()

            # prepend in-context cls token
            _, D = cos_img.shape
            cos_pad = torch.ones(num_cls_token, D, dtype=cos_img.dtype, device=cos_img.device)
            sin_pad = torch.zeros(num_cls_token, D, dtype=sin_img.dtype, device=sin_img.device)

            freqs_cos = torch.cat([cos_pad, cos_img], dim=0)  # [N_cls+N_img, D]
            freqs_sin = torch.cat([sin_pad, sin_img], dim=0)
        else:
            freqs_cos = freqs.cos().view(-1, freqs.shape[-1])
            freqs_sin = freqs.sin().view(-1, freqs.shape[-1])

        self.register_buffer("freqs_cos", freqs_cos, persistent=False)
        self.register_buffer("freqs_sin", freqs_sin, persistent=False)

    def forward(self, t):
        return t * self.freqs_cos + rotate_half(t) * self.freqs_sin


class RMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6) -> None:
        """
        LlamaRMSNorm is equivalent to T5LayerNorm
        """
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps

    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return (self.weight * hidden_states).to(input_dtype)


def get_2d_sincos_pos_embed(embed_dim, grid_size, cls_token=False, extra_tokens=0):
    """
    grid_size: int of the grid height and width
    return:
    pos_embed: [grid_size*grid_size, embed_dim] or [1+grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    grid_h = np.arange(grid_size, dtype=np.float32)
    grid_w = np.arange(grid_size, dtype=np.float32)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size, grid_size])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token and extra_tokens > 0:
        pos_embed = np.concatenate([np.zeros([extra_tokens, embed_dim]), pos_embed], axis=0)
    return pos_embed


def get_2d_sincos_pos_embed_from_grid(embed_dim, grid):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[0])  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[1])  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1)  # (H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float64)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb
